#
# load a few simple objects
#
source objects/clamp.tcl
source objects/cube.tcl
source objects/arrow.tcl

#
# cport:
#
# create a canvas viewport item in a toplevel window with the
# specified colormap and visual, and associate the viewport item
# with the given cell.
#

proc cport {cell top cmap visual args} {
 
    catch {destroy $top};
    toplevel $top -bd 0 -colormap $cmap -visual $visual;
    wm geometry $top 400x320;

    frame $top.f -relief flat -bd 0;
    button $top.f.done -text "DONE" -width 10 -relief raised -bd 2 \
		       -command "destroy $top; $cell destroy";
    pack $top.f.done -anchor center -side right;

    button $top.f.spin -text "SPIN" -width 10 -relief raised -bd 2 \
		       -command "look-at $top.c .1 {0 0 0}";
    pack $top.f.spin -anchor center -side left;
    pack $top.f -side bottom -fill x -padx 20 -pady 5;

    canvas $top.c;
    pack $top.c -expand 1 -fill both -padx 10 -pady 10;
    $top.c configure -bg steelblue;
  
    set id [$top.c create viewport 30 20 350 260 \
	    -minu -50 -minv -50 -maxu 50 -maxv 50 \
	    -fill black \
	    -outline red \
	    -projection perspective \
	    -vrp {0 0 100} -prp {0 0 100} \
	    -vpn {0 0.2 1} -vup {0 1 0} \
	    -ambient {#5fff5fff5fff} \
	    -depthcue {black linear 0 1000} \
	    -light0 {white white white {50 50 50 0} {-1 -1 -1 } 1 70 {1 0 0}} \
	    ];

    $top.c itemconfigure $id -cell $cell;
    if {$args != ""} {eval [concat "$top.c itemconfigure $id" $args]};
    update;

    $top.c bind $id <Button-1> "puts \[list %x %y\]; \
			        $top.c itemconfigure $id -pick \[list %x %y 1 1\]; \
			        puts \[$top.c itemcget $id -pick\]";
    $top.c bind $id <Button-2> "set $cell\_stop_rotation 1";
}


#
# vport:
#
# create a viewport widget in a toplevel window with the
# specified colormap and visual, and associate the viewport widget
# with the given cell.
#

proc vport {cell top cmap visual args} {

    catch {destroy $top};
    toplevel $top -relief sunken -bd 0 -colormap $cmap -visual $visual;
    wm geometry $top 400x320;

    frame $top.f -relief flat -bd 0;
    button $top.f.done -text "DONE" -width 10 -relief raised -bd 2 \
		       -command "destroy $top; $cell destroy";
    pack $top.f.done -anchor center -side right;

    button $top.f.spin -text "SPIN" -width 10 -relief raised -bd 2 \
		       -command "look-at $top.c .1 {0 0 0}";
    pack $top.f.spin -anchor center -side left;
    pack $top.f -side bottom -fill x -padx 20 -pady 5;

    viewport $top.c 1 1;
    pack $top.c -expand 1 -fill both -padx 10 -pady 10;
  
    $top.c configure \
      -minu -50 -minv -50 -maxu 50 -maxv 50 \
      -fill black \
      -outline red -width 1 \
      -projection perspective \
      -vrp {0 0 100} -prp {0 0 100} \
      -vpn {0.1 0 1} -vup {0 1 0} \
      -ambient {#5fff5fff5fff} \
      -depthcue {black linear 0 1000} \
      -light0 {white white white {50 50 50 0} {-1 -1 -1 } 1 45 {1 0 0}} \
      -light1 {white white white {-50 50 50 0} {1 -1 -1 } 1 10 {1 0 0}};

    $top.c config -cell $cell;
    if {$args != ""} {eval [concat "$top.c config " $args]};
    update;

    bind $top.c <Button-1> "puts \[list %x %y\]; \
			    $top.c configure -pick \[list %x %y 1 1\]; \
			    puts \[$top.c cget -pick\]";
    bind $top.c <Button-2> "set $cell\_stop_rotation 1";
}

#
# populate:
#
# populate a cell with a few simple objects:
#

proc populate {cell} {

    if ![catch {cell create $cell}] {

	mkclamp $cell clamp1 \
	  -pos {5 10 0} \
	  -fwd {1 0 1} -up {0 1 0} \
	  -shading smooth \
	  -color red \
	  -emission {0 0 0} \
	  -ambient {0.4 0.4 0.4} \
	  -diffuse {1 1 1} \
	  -spec {white 15} \
	  -average on;

	mkcube $cell cube1 10 \
	  -pos {25 10 25} \
	  -shading smooth \
	  -color white \
	  -emission {0 0 0} \
	  -ambient {0.4 0.4 0.4} \
	  -diffuse {1 1 1} \
	  -spec {white 15} \
	  -average off \
	  -parent clamp1;

	$cell create sphere sphere1 \
	  -radius 10 \
	  -slices 20 \
	  -stacks 20 \
	  -pos {-20 -15 -20} \
	  -fwd {0 0 1} -up {0 1 0} \
	  -shading smooth \
	  -normals smooth \
	  -color green \
	  -emission {0.8 0.8 0.8} \
	  -ambient {0 0 0} \
	  -diffuse {0.5 0.5 0.5} \
	  -spec {white 20} \
	  -parent clamp1;
	
	$cell create cylinder cyl1 \
	  -topradius 0 \
	  -baseradius 10 \
	  -height 20 \
	  -pos {-40 -50 -40} \
	  -fwd {0 1 0} -up {0 0 1} \
	  -shading smooth \
	  -normals smooth \
	  -color blue \
	  -emission {0 0 0} \
	  -ambient {0.4 0.4 0.4} \
	  -diffuse {1 1 1} \
	  -spec {white 15} \
	  -parent clamp1;
    }
    return [$cell list];
}

#
# rotate-y:
#
# rotate objects in a cell clockwise about the y-axis
# one revolution in steps of $dt
#

proc rotate-y {c dt objects} {

    upvar #0 $c\_stop_rotation stop;

    set stop 0;
    set twopi [expr 2.0 * 3.1415927];
    set cosine [expr cos($dt)];
    set sine [expr sin($dt)];

    for {set theta 0.0} {$theta < $twopi} {set theta [expr $theta + $dt]} {

	foreach obj $objects {
	    if {$stop == 1} {return}

	    set fwd [$c modelcget $obj -fwd];
	    set up [$c modelcget $obj -up];

	    set fwdx [lindex $fwd 0];
	    set fwdy [lindex $fwd 1];
	    set fwdz [lindex $fwd 2];

	    set upx [lindex $up 0];
	    set upy [lindex $up 1];
	    set upz [lindex $up 2];

	    set nfwdx [expr $fwdx * $cosine + $fwdz * $sine];
	    set nfwdz [expr -$fwdx * $sine + $fwdz * $cosine];

	    set nupx [expr $upx * $cosine + $upz * $sine];
	    set nupz [expr -$upx * $sine + $upz * $cosine];

	    $c config $obj -fwd [list $nfwdx $fwdy $nfwdz] \
			   -up [list $nupx $upy $nupz];
	}
	update;
    }
}

#
# rotate-x:
#
# rotate objects in a cell clockwise about the x-axis
# one revolution in steps of $dt
#

proc rotate-x {c dt objects} {

    upvar #0 $c\_stop_rotation stop;

    set stop 0;
    set twopi [expr 2.0 * 3.1415927];
    set cosine [expr cos($dt)];
    set sine [expr sin($dt)];

    for {set theta 0.0} {$theta < $twopi} {set theta [expr $theta + $dt]} {

	foreach obj $objects {

	    if {$stop == 1} {return}

	    set fwd [$c modelcget $obj -fwd];
	    set up [$c modelcget $obj -up];

	    set fwdx [lindex $fwd 0];
	    set fwdy [lindex $fwd 1];
	    set fwdz [lindex $fwd 2];

	    set upx [lindex $up 0];
	    set upy [lindex $up 1];
	    set upz [lindex $up 2];

	    set nfwdy [expr $fwdy * $cosine - $fwdz * $sine];
	    set nfwdz [expr $fwdy * $sine + $fwdz * $cosine];

	    set nupy [expr $upy * $cosine - $upz * $sine];
	    set nupz [expr $upy * $sine + $upz * $cosine];
	  
	    $c config $obj -fwd [list $fwdx $nfwdy $nfwdz] \
			   -up [list $upx $nupy $nupz];
	}
	update;
    }
}

#
# rotate-z:
#
# rotate objects in a cell clockwise about the z-axis
# one revolution in steps of $dt
#

proc rotate-z {c dt args} {

    upvar #0 $c\_stop_rotation stop;

    set stop 0;
    set twopi [expr 2.0 * 3.1415927];
    set cosine [expr cos($dt)];
    set sine [expr sin($dt)];

    for {set theta 0.0} {$theta < $twopi} {set theta [expr $theta + $dt]} {

	foreach obj $args {

	    if {$stop == 1} {return};

	    set fwd [$c modelcget $obj -fwd];
	    set up [$c modelcget $obj -up];

	    set fwdx [lindex $fwd 0];
	    set fwdy [lindex $fwd 1];
	    set fwdz [lindex $fwd 2];

	    set upx [lindex $up 0];
	    set upy [lindex $up 1];
	    set upz [lindex $up 2];

	    set nfwdy [expr $fwdy * $cosine + $fwdx * $sine];
	    set nfwdx [expr -$fwdy * $sine + $fwdx * $cosine];

	    set nupy [expr $upy * $cosine + $upx * $sine];
	    set nupx [expr -$upy * $sine + $upx * $cosine];

	    $c modelconfigure $obj -fwd [list $nfwdx $nfwdy $fwdz] \
				   -up [list $nupx $nupy $upz];
	}
	update;
    }
}

#
# look-at:
#
# look at a specific location in space from all angles


proc look-at {port dt pos} {

  set cell [$port cget -cell];

  upvar #0 $cell\_stop_rotation stop;

  set stop 0;
  set twopi [expr 2.0 * 3.1415927];
  set cosine [expr cos($dt)];
  set sine [expr sin($dt)];

  set vrp [$port cget -vrp];

  set vrpx [lindex $vrp 0];
  set vrpy [lindex $vrp 1];
  set vrpz [lindex $vrp 2];

  set posx [lindex $pos 0];
  set posy [lindex $pos 1];
  set posz [lindex $pos 2];

  set vpnx [expr $vrpx - $posx];
  set vpny [expr $vrpy - $posy];
  set vpnz [expr $vrpz - $posz];

  if {($vpnx == 0) && ($vpnz == 0)} {
    puts "view projection normal cannot be parallel to the y-axis";
    return;
  }
  $port configure -vpn [list $vpnx $vpny $vpnz] \
		  -vup {0 1 0};

  for {set theta 0.0} {$theta < $twopi} {set theta [expr $theta + $dt]} {

    if ![winfo exists $port] {return};
    if {$stop == 1} {return};

    set nvrpx [expr $vrpx * $cosine + $vrpz * $sine];
    set nvrpz [expr -$vrpx * $sine + $vrpz * $cosine];

    $port config -vrp [list $nvrpx $vrpy $nvrpz] \
		 -vpn [list $nvrpx $vrpy $nvrpz];

    set vrpx $nvrpx;
    set vrpz $nvrpz;

    update;
  }
}
