/* 
 * smoption.c --
 *
 *	This file implements model object options.
 *
 */

#include "copyright.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <tk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "smInt.h"


/*
 * Prototypes for procedures defined externally:
 */

void		SmRemoveObjectFromParent _ANSI_ARGS_((Object *obj));

Object		*Sm_FindObject _ANSI_ARGS_((Cell *cell, char *label));

int		SmVectorParseProc _ANSI_ARGS_((ClientData clientData,
						Tcl_Interp *interp, Tk_Window tkwin,
						char *value, char *widgRec, int offset));

char		*SmVectorPrintProc _ANSI_ARGS_((ClientData clientData,
						 Tk_Window tkwin, char *widgRec, int offset,
						 Tcl_FreeProc **freeProcPtr));


/*
 * Prototypes for procedures defined in this file that are exported:
 */


/*
 * Prototypes for procedures defined in this file that are static:
 */

static int	ObjectParentParseProc _ANSI_ARGS_((ClientData clientData,
						   Tcl_Interp *interp, Tk_Window tkwin,
						   char *value, char *widgRec, int offset));

static char *	ObjectParentPrintProc _ANSI_ARGS_((ClientData clientData,
						   Tk_Window tkwin, char *widgRec, int offset,
						   Tcl_FreeProc **freeProcPtr));

static int	ObjectVerticesParseProc _ANSI_ARGS_((ClientData clientData,
						     Tcl_Interp *interp, Tk_Window tkwin,
						     char *value, char *widgRec, int offset));

static char *	ObjectVerticesPrintProc _ANSI_ARGS_((ClientData clientData,
						     Tk_Window tkwin, char *widgRec, int offset,
						     Tcl_FreeProc **freeProcPtr));

static int	ObjectSurfacesParseProc _ANSI_ARGS_((ClientData clientData,
						     Tcl_Interp *interp, Tk_Window tkwin,
						     char *value, char *widgRec, int offset));

static char *	ObjectSurfacesPrintProc _ANSI_ARGS_((ClientData clientData,
						     Tk_Window tkwin, char *widgRec, int offset,
						     Tcl_FreeProc **freeProcPtr));

static int	ObjectNormalsParseProc _ANSI_ARGS_((ClientData clientData,
						    Tcl_Interp *interp, Tk_Window tkwin,
						    char *value, char *widgRec, int offset));

static char *	ObjectNormalsPrintProc _ANSI_ARGS_((ClientData clientData,
						    Tk_Window tkwin, char *widgRec, int offset,
						    Tcl_FreeProc **freeProcPtr));

static int	ObjectShadingParseProc _ANSI_ARGS_((ClientData clientData,
						    Tcl_Interp *interp, Tk_Window tkwin,
						    char *value, char *widgRec, int offset));

static char *	ObjectShadingPrintProc _ANSI_ARGS_((ClientData clientData,
						    Tk_Window tkwin, char *widgRec, int offset,
						    Tcl_FreeProc **freeProcPtr));

static int	ObjectCoefficientParseProc _ANSI_ARGS_((ClientData clientData,
							Tcl_Interp *interp, Tk_Window tkwin,
							char *value, char *widgRec, int offset));

static char *	ObjectCoefficientPrintProc _ANSI_ARGS_((ClientData clientData,
							Tk_Window tkwin, char *widgRec, int offset,
							Tcl_FreeProc **freeProcPtr));

static int	ObjectSpecularParseProc _ANSI_ARGS_((ClientData clientData,
						     Tcl_Interp *interp, Tk_Window tkwin,
						     char *value, char *widgRec, int offset));

static char *	ObjectSpecularPrintProc _ANSI_ARGS_((ClientData clientData,
						     Tk_Window tkwin, char *widgRec, int offset,
						     Tcl_FreeProc **freeProcPtr));

static int	PolygonModeParseProc _ANSI_ARGS_((ClientData clientData,
						  Tcl_Interp *interp, Tk_Window tkwin,
						  char *value, char *widgRec, int offset));

static char *	PolygonModePrintProc _ANSI_ARGS_((ClientData clientData,
						  Tk_Window tkwin, char *widgRec, int offset,
						  Tcl_FreeProc **freeProcPtr));

static int	PolygonCullFaceParseProc _ANSI_ARGS_((ClientData clientData,
						      Tcl_Interp *interp, Tk_Window tkwin,
						      char *value, char *widgRec, int offset));

static char *	PolygonCullFacePrintProc _ANSI_ARGS_((ClientData clientData,
						      Tk_Window tkwin, char *widgRec, int offset,
						      Tcl_FreeProc **freeProcPtr));

static int	QuadricDrawStyleParseProc _ANSI_ARGS_((ClientData clientData,
						       Tcl_Interp *interp, Tk_Window tkwin,
						       char *value, char *widgRec, int offset));

static char *	QuadricDrawStylePrintProc _ANSI_ARGS_((ClientData clientData,
						       Tk_Window tkwin, char *widgRec, int offset,
						       Tcl_FreeProc **freeProcPtr));

static int	QuadricNormalsParseProc _ANSI_ARGS_((ClientData clientData,
						     Tcl_Interp *interp, Tk_Window tkwin,
						     char *value, char *widgRec, int offset));

static char *	QuadricNormalsPrintProc _ANSI_ARGS_((ClientData clientData,
						     Tk_Window tkwin, char *widgRec, int offset,
						     Tcl_FreeProc **freeProcPtr));

static int	QuadricOrientationParseProc _ANSI_ARGS_((ClientData clientData,
							 Tcl_Interp *interp, Tk_Window tkwin,
							 char *value, char *widgRec, int offset));

static char *	QuadricOrientationPrintProc _ANSI_ARGS_((ClientData clientData,
							 Tk_Window tkwin, char *widgRec, int offset,
							 Tcl_FreeProc **freeProcPtr));


/*
 * Custom options for cell objects
 */

static Tk_CustomOption sm_ObjectOptionCULL = {
    PolygonCullFaceParseProc,
    PolygonCullFacePrintProc,
    (ClientData) NULL,
};

static Tk_CustomOption sm_ObjectOptionFRONT = {
    PolygonModeParseProc,
    PolygonModePrintProc,
    (ClientData) GL_FRONT,
};

static Tk_CustomOption sm_ObjectOptionBACK = {
    PolygonModeParseProc,
    PolygonModePrintProc,
    (ClientData) GL_BACK,
};

static Tk_CustomOption sm_ObjectOptionFWD = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_OBJECT_FWD,
};

static Tk_CustomOption sm_ObjectOptionUP = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_OBJECT_UP,
};

static Tk_CustomOption sm_ObjectOptionPOS = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_OBJECT_POS,
};

static Tk_CustomOption sm_ObjectOptionParent = {
    ObjectParentParseProc,
    ObjectParentPrintProc,
    NULL,
};

static Tk_CustomOption sm_ObjectOptionVertices = {
    ObjectVerticesParseProc,
    ObjectVerticesPrintProc,
    NULL,
};

static Tk_CustomOption sm_ObjectOptionSurfaces = {
    ObjectSurfacesParseProc,
    ObjectSurfacesPrintProc,
    NULL,
};

static Tk_CustomOption sm_ObjectOptionNormals = {
    ObjectNormalsParseProc,
    ObjectNormalsPrintProc,
    NULL,
};

static Tk_CustomOption sm_ObjectOptionShading = {
    ObjectShadingParseProc,
    ObjectShadingPrintProc,
    NULL,
};

static Tk_CustomOption sm_ObjectOptionAmbient = {
    ObjectCoefficientParseProc,
    ObjectCoefficientPrintProc,
    (ClientData) SM_OBJECT_AMBIENT,
};

static Tk_CustomOption sm_ObjectOptionDiffuse = {
    ObjectCoefficientParseProc,
    ObjectCoefficientPrintProc,
    (ClientData) SM_OBJECT_DIFFUSE,
};

static Tk_CustomOption sm_ObjectOptionEmission = {
    ObjectCoefficientParseProc,
    ObjectCoefficientPrintProc,
    (ClientData) SM_OBJECT_EMISSION,
};

static Tk_CustomOption sm_ObjectOptionSpecular = {
    ObjectSpecularParseProc,
    ObjectSpecularPrintProc,
    (ClientData) SM_OBJECT_SPECULAR,
};

static Tk_CustomOption sm_ObjectOptionBFAmbient = {
    ObjectCoefficientParseProc,
    ObjectCoefficientPrintProc,
    (ClientData) SM_OBJECT_BF_AMBIENT,
};

static Tk_CustomOption sm_ObjectOptionBFDiffuse = {
    ObjectCoefficientParseProc,
    ObjectCoefficientPrintProc,
    (ClientData) SM_OBJECT_BF_DIFFUSE,
};

static Tk_CustomOption sm_ObjectOptionBFEmission = {
    ObjectCoefficientParseProc,
    ObjectCoefficientPrintProc,
    (ClientData) SM_OBJECT_BF_EMISSION,
};

static Tk_CustomOption sm_ObjectOptionBFSpecular = {
    ObjectSpecularParseProc,
    ObjectSpecularPrintProc,
    (ClientData) SM_OBJECT_BF_SPECULAR,
};

static Tk_CustomOption sm_QuadricOptionDrawStyle = {
    QuadricDrawStyleParseProc,
    QuadricDrawStylePrintProc,
    NULL,
};

static Tk_CustomOption sm_QuadricOptionNormals = {
    QuadricNormalsParseProc,
    QuadricNormalsPrintProc,
    NULL,
};

static Tk_CustomOption sm_QuadricOptionOrientation = {
    QuadricOrientationParseProc,
    QuadricOrientationPrintProc,
    NULL,
};


/* Shading types */
static char *surfaceShadingMethods[] = {"flat", "smooth", (char *) NULL};


/*
 * Information used for parsing polygon object configuration specs:
 */

Tk_ConfigSpec objectPolygonConfigSpecs[] = {

    {TK_CONFIG_CUSTOM, "-ambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionAmbient},

    {TK_CONFIG_BOOLEAN, "-average", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, useAverage), 0},

    {TK_CONFIG_CUSTOM, "-backmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionBACK},

    {TK_CONFIG_CUSTOM, "-bfambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionBFAmbient},

    {TK_CONFIG_COLOR, "-bfcolor", "surfaceColor", "SurfaceColor",
       "black", Tk_Offset(Object, bfcolor), },

    {TK_CONFIG_CUSTOM, "-bfdiffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionBFDiffuse},

    {TK_CONFIG_CUSTOM, "-bfemission", "emissionCoefficient", "EmissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionBFEmission},

    {TK_CONFIG_CUSTOM, "-bfspecular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionBFSpecular},

    {TK_CONFIG_BOOLEAN, "-ccw", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, ccw), 0},

    {TK_CONFIG_COLOR, "-color", "surfaceColor", "SurfaceColor",
       "white", Tk_Offset(Object, color), },

    {TK_CONFIG_CUSTOM, "-cull", (char *) NULL, (char *) NULL,
       "back", 0, 0, &sm_ObjectOptionCULL},

    {TK_CONFIG_CUSTOM, "-diffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionDiffuse},

    {TK_CONFIG_CUSTOM, "-emission", "emissionCoefficient", "emissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionEmission},

    {TK_CONFIG_CUSTOM, "-frontmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionFRONT},

    {TK_CONFIG_CUSTOM, "-fwd", (char *) NULL, (char *) NULL,
       "0.0 0.0 1.0", 0, 0, &sm_ObjectOptionFWD},

    {TK_CONFIG_CUSTOM, "-normals", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionNormals},

    {TK_CONFIG_CUSTOM, "-parent", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionParent},

    {TK_CONFIG_CUSTOM, "-position", (char *) NULL, (char *) NULL,
       "0.0 0.0 0.0", 0, 0, &sm_ObjectOptionPOS},

    {TK_CONFIG_CUSTOM, "-shading", (char *) NULL, (char *) NULL,
       "flat", 0, 0, &sm_ObjectOptionShading},

    {TK_CONFIG_CUSTOM, "-specular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionSpecular},

    {TK_CONFIG_CUSTOM, "-surfaces", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionSurfaces},

    {TK_CONFIG_CUSTOM, "-up", (char *) NULL, (char *) NULL,
       "0.0 1.0 0.0", 0, 0, &sm_ObjectOptionUP},

    {TK_CONFIG_CUSTOM, "-vertices", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionVertices},

    {TK_CONFIG_BOOLEAN, "-visible", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, visible), 0},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};

/*
 * Information used for parsing cylinder object configuration specs:
 */

Tk_ConfigSpec objectCylinderConfigSpecs[] = {

    {TK_CONFIG_CUSTOM, "-ambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionAmbient},

    {TK_CONFIG_CUSTOM, "-backmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionBACK},

    {TK_CONFIG_DOUBLE, "-baseradius", (char *) NULL, (char *) NULL,
       "1", Tk_Offset(Object, baseRadius), 0},

    {TK_CONFIG_CUSTOM, "-bfambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionBFAmbient},

    {TK_CONFIG_COLOR, "-bfcolor", "surfaceColor", "SurfaceColor",
       "black", Tk_Offset(Object, bfcolor), },

    {TK_CONFIG_CUSTOM, "-bfdiffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionBFDiffuse},

    {TK_CONFIG_CUSTOM, "-bfemission", "emissionCoefficient", "EmissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionBFEmission},

    {TK_CONFIG_CUSTOM, "-bfspecular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionBFSpecular},

    {TK_CONFIG_BOOLEAN, "-ccw", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, ccw), 0},

    {TK_CONFIG_COLOR, "-color", "surfaceColor", "SurfaceColor",
       "white", Tk_Offset(Object, color), },

    {TK_CONFIG_CUSTOM, "-cull", (char *) NULL, (char *) NULL,
       "none", 0, 0, &sm_ObjectOptionCULL},

    {TK_CONFIG_CUSTOM, "-diffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionDiffuse},

    {TK_CONFIG_CUSTOM, "-drawstyle", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_QuadricOptionDrawStyle},

    {TK_CONFIG_CUSTOM, "-emission", "emissionCoefficient", "emissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionEmission},

    {TK_CONFIG_CUSTOM, "-frontmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionFRONT},

    {TK_CONFIG_CUSTOM, "-fwd", (char *) NULL, (char *) NULL,
       "0.0 0.0 1.0", 0, 0, &sm_ObjectOptionFWD},

    {TK_CONFIG_DOUBLE, "-height", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, height), 0},

    {TK_CONFIG_CUSTOM, "-normals", (char *) NULL, (char *) NULL,
       "none", 0, 0, &sm_QuadricOptionNormals},

    {TK_CONFIG_CUSTOM, "-orientation", (char *) NULL, (char *) NULL,
       "outside", 0, 0, &sm_QuadricOptionOrientation},

    {TK_CONFIG_CUSTOM, "-parent", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionParent},

    {TK_CONFIG_CUSTOM, "-position", (char *) NULL, (char *) NULL,
       "0.0 0.0 0.0", 0, 0, &sm_ObjectOptionPOS},

    {TK_CONFIG_CUSTOM, "-shading", (char *) NULL, (char *) NULL,
       "flat", 0, 0, &sm_ObjectOptionShading},

    {TK_CONFIG_INT, "-slices", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, slices), 0},

    {TK_CONFIG_CUSTOM, "-specular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionSpecular},

    {TK_CONFIG_INT, "-stacks", (char *) NULL, (char *) NULL,
       "1", Tk_Offset(Object, stacks), 0},

    {TK_CONFIG_DOUBLE, "-topradius", (char *) NULL, (char *) NULL,
       "1", Tk_Offset(Object, topRadius), 0},

    {TK_CONFIG_CUSTOM, "-up", (char *) NULL, (char *) NULL,
       "0.0 1.0 0.0", 0, 0, &sm_ObjectOptionUP},

    {TK_CONFIG_BOOLEAN, "-visible", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, visible), 0},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};

/*
 * Information used for parsing disk object configuration specs:
 */

Tk_ConfigSpec objectDiskConfigSpecs[] = {

    {TK_CONFIG_CUSTOM, "-ambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionAmbient},

    {TK_CONFIG_CUSTOM, "-backmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionBACK},

    {TK_CONFIG_CUSTOM, "-bfambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionBFAmbient},

    {TK_CONFIG_COLOR, "-bfcolor", "surfaceColor", "SurfaceColor",
       "black", Tk_Offset(Object, bfcolor), },

    {TK_CONFIG_CUSTOM, "-bfdiffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionBFDiffuse},

    {TK_CONFIG_CUSTOM, "-bfemission", "emissionCoefficient", "EmissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionBFEmission},

    {TK_CONFIG_CUSTOM, "-bfspecular", "specularReflection", "SpecularReflection",
       "white 1.0", 0, 0, &sm_ObjectOptionBFSpecular},

    {TK_CONFIG_BOOLEAN, "-ccw", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, ccw), 0},

    {TK_CONFIG_COLOR, "-color", "surfaceColor", "SurfaceColor",
       "white", Tk_Offset(Object, color), },

    {TK_CONFIG_CUSTOM, "-cull", (char *) NULL, (char *) NULL,
       "none", 0, 0, &sm_ObjectOptionCULL},

    {TK_CONFIG_CUSTOM, "-diffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionDiffuse},

    {TK_CONFIG_CUSTOM, "-drawstyle", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_QuadricOptionDrawStyle},

    {TK_CONFIG_CUSTOM, "-emission", "emissionCoefficient", "emissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionEmission},

    {TK_CONFIG_CUSTOM, "-frontmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionFRONT},

    {TK_CONFIG_CUSTOM, "-fwd", (char *) NULL, (char *) NULL,
       "0.0 0.0 1.0", 0, 0, &sm_ObjectOptionFWD},

    {TK_CONFIG_DOUBLE, "-innerradius", (char *) NULL, (char *) NULL,
       "5", Tk_Offset(Object, innerRadius), 0},

    {TK_CONFIG_INT, "-loops", (char *) NULL, (char *) NULL,
       "1", Tk_Offset(Object, loops), 0},

    {TK_CONFIG_CUSTOM, "-normals", (char *) NULL, (char *) NULL,
       "none", 0, 0, &sm_QuadricOptionNormals},

    {TK_CONFIG_CUSTOM, "-orientation", (char *) NULL, (char *) NULL,
       "outside", 0, 0, &sm_QuadricOptionOrientation},

    {TK_CONFIG_DOUBLE, "-outerradius", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, outerRadius), 0},

    {TK_CONFIG_CUSTOM, "-parent", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionParent},

    {TK_CONFIG_CUSTOM, "-position", (char *) NULL, (char *) NULL,
       "0.0 0.0 0.0", 0, 0, &sm_ObjectOptionPOS},

    {TK_CONFIG_CUSTOM, "-shading", (char *) NULL, (char *) NULL,
       "flat", 0, 0, &sm_ObjectOptionShading},

    {TK_CONFIG_INT, "-slices", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, slices), 0},

    {TK_CONFIG_CUSTOM, "-specular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionSpecular},

    {TK_CONFIG_CUSTOM, "-up", (char *) NULL, (char *) NULL,
       "0.0 1.0 0.0", 0, 0, &sm_ObjectOptionUP},

    {TK_CONFIG_BOOLEAN, "-visible", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, visible), 0},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};

/*
 * Information used for parsing sphere object configuration specs:
 */

Tk_ConfigSpec objectSphereConfigSpecs[] = {

    {TK_CONFIG_CUSTOM, "-ambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionAmbient},

    {TK_CONFIG_CUSTOM, "-backmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionBACK},

    {TK_CONFIG_CUSTOM, "-bfambient", "ambientCoefficient", "AmbientCoefficient",
       "0.2 0.2 0.2", 0, 0, &sm_ObjectOptionBFAmbient},

    {TK_CONFIG_COLOR, "-bfcolor", "surfaceColor", "SurfaceColor",
       "black", Tk_Offset(Object, bfcolor), },

    {TK_CONFIG_CUSTOM, "-bfdiffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionBFDiffuse},

    {TK_CONFIG_CUSTOM, "-bfemission", "emissionCoefficient", "EmissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionBFEmission},

    {TK_CONFIG_CUSTOM, "-bfspecular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionBFSpecular},

    {TK_CONFIG_BOOLEAN, "-ccw", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, ccw), 0},

    {TK_CONFIG_COLOR, "-color", "surfaceColor", "SurfaceColor",
       "white", Tk_Offset(Object, color), },

    {TK_CONFIG_CUSTOM, "-cull", (char *) NULL, (char *) NULL,
       "none", 0, 0, &sm_ObjectOptionCULL},

    {TK_CONFIG_CUSTOM, "-diffuse", "diffuseCoefficient", "DiffuseCoefficient",
       "0.8 0.8 0.8", 0, 0, &sm_ObjectOptionDiffuse},

    {TK_CONFIG_CUSTOM, "-drawstyle", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_QuadricOptionDrawStyle},

    {TK_CONFIG_CUSTOM, "-emission", "emissionCoefficient", "emissionCoefficient",
       "0 0 0", 0, 0, &sm_ObjectOptionEmission},

    {TK_CONFIG_CUSTOM, "-frontmode", (char *) NULL, (char *) NULL,
       "fill", 0, 0, &sm_ObjectOptionFRONT},

    {TK_CONFIG_CUSTOM, "-fwd", (char *) NULL, (char *) NULL,
       "0.0 0.0 1.0", 0, 0, &sm_ObjectOptionFWD},

    {TK_CONFIG_CUSTOM, "-normals", (char *) NULL, (char *) NULL,
       "none", 0, 0, &sm_QuadricOptionNormals},

    {TK_CONFIG_CUSTOM, "-orientation", (char *) NULL, (char *) NULL,
       "outside", 0, 0, &sm_QuadricOptionOrientation},

    {TK_CONFIG_CUSTOM, "-parent", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &sm_ObjectOptionParent},

    {TK_CONFIG_CUSTOM, "-position", (char *) NULL, (char *) NULL,
       "0.0 0.0 0.0", 0, 0, &sm_ObjectOptionPOS},

    {TK_CONFIG_DOUBLE, "-radius", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, radius), 0},

    {TK_CONFIG_CUSTOM, "-shading", (char *) NULL, (char *) NULL,
       "flat", 0, 0, &sm_ObjectOptionShading},

    {TK_CONFIG_INT, "-slices", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, slices), 0},

    {TK_CONFIG_CUSTOM, "-specular", "specularReflection", "SpecularReflection",
       "black 0.0", 0, 0, &sm_ObjectOptionSpecular},

    {TK_CONFIG_INT, "-stacks", (char *) NULL, (char *) NULL,
       "10", Tk_Offset(Object, stacks), 0},

    {TK_CONFIG_CUSTOM, "-up", (char *) NULL, (char *) NULL,
       "0.0 1.0 0.0", 0, 0, &sm_ObjectOptionUP},

    {TK_CONFIG_BOOLEAN, "-visible", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(Object, visible), 0},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};



/*
 *--------------------------------------------------------------
 *
 * ObjectParentParseProc:
 *
 * process the '-parent' option of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectParentParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;
    Tcl_Interp *interp;
    Tk_Window tkwin;
    char *value;
    char *widgRec;
    int offset;
{
    Object *p, *obj = (Object *) widgRec;

    if (*value) {
	if ((p = Sm_FindObject(obj->cell, value)) == NULL) {
	    Tcl_AppendResult(interp, "object \"", value, "\" not found.", (char *) NULL);
	    return TCL_ERROR;
	}
	SmRemoveObjectFromParent(obj);
	obj->parent = p;
	obj->sibling = p->child;
	p->child = obj;
    } else {
	SmRemoveObjectFromParent(obj);
	obj->parent = NULL;
    }
    obj->updateFlag |= SM_OBJECT_UPDATE_POSITION;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectParentPrintProc:
 *
 * produce a printable string for the '-parent' option.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectParentPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *p = ((Object *) widgRec)->parent;

    *freeProcPtr = NULL;

    if (p) {
	return p->label;
    } else {
	return "";
    }
}

/*
 *--------------------------------------------------------------
 *
 * ObjectVerticesParseProc:
 *
 * process the '-vertices' option of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectVerticesParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;
    Tcl_Interp *interp;
    Tk_Window tkwin;
    char *value;
    char *widgRec;
    int offset;
{
    int i, argc, nc;
    char **argv, **comp;
    double lv0, lv1, lv2;
    Object *obj = (Object *) widgRec;
    Vertex *v;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }

    if (argc == 0) {
	if (obj->v) {
	    free((void *) obj->v);
	}
	obj->nv = 0;
	free((void *) argv);
	return TCL_OK;
    }

    if ((v = (Vertex *) malloc(argc * sizeof(Vertex))) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for vertex list.", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }

    for (i = 0; i < argc; i++) {

	if (Tcl_SplitList(interp, argv[i], &nc, &comp) != TCL_OK) {
	    free((void *) argv);
	    free((void *) v);
	    return TCL_ERROR;
	}

	if (nc != 3) {
	    Tcl_AppendResult(interp, "bad vertex value \"", argv[i], "\": three components expected.", (char *) NULL);
	    free((void *) comp);
	    free((void *) argv);
	    free((void *) v);
	    return TCL_ERROR;
	}

	if (Tcl_GetDouble(interp, comp[0], &lv0) != TCL_OK ||
	    Tcl_GetDouble(interp, comp[1], &lv1) != TCL_OK ||
	    Tcl_GetDouble(interp, comp[2], &lv2) != TCL_OK) {
	    free((void *) comp);
	    free((void *) argv);
	    free((void *) v);
	    return TCL_ERROR;
	}
	v[i].lv[0] = (DOUBLE) lv0;
	v[i].lv[1] = (DOUBLE) lv1;
	v[i].lv[2] = (DOUBLE) lv2;
	free((void *) comp);
    }
    obj->nv = argc;
    if (obj->v) free((void *) obj->v);
    free((void *) argv);
    obj->v = v;
    obj->updateFlag |= SM_OBJECT_UPDATE_NORMAL + SM_OBJECT_UPDATE_POSITION;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectVerticesPrintProc:
 *
 * produce a printable string for the '-vertices' option.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectVerticesPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    int i;
    char *buf;
    Vertex *v;
    Object *obj = (Object *) widgRec;
    Tcl_Interp *interp = obj->cell->interp;
    char c0[TCL_DOUBLE_SPACE], c1[TCL_DOUBLE_SPACE], c2[TCL_DOUBLE_SPACE];

    if (obj->nv == 0) {
	*freeProcPtr = NULL;
	return "";
    }

    if ((buf = (char *) malloc(obj->nv * 80)) == NULL) {
	*freeProcPtr = NULL;
	return "";
    }

    v = obj->v;
    *buf = '\0';
    for (i = 0; i < obj->nv - 1; i++) {
	Tcl_PrintDouble(interp, (double) v[i].lv[0], c0);
	Tcl_PrintDouble(interp, (double) v[i].lv[1], c1);
	Tcl_PrintDouble(interp, (double) v[i].lv[2], c2);
	sprintf(buf + strlen(buf), "{%s %s %s} ", c0, c1, c2);
    }
    Tcl_PrintDouble(interp, (double) v[i].lv[0], c0);
    Tcl_PrintDouble(interp, (double) v[i].lv[1], c1);
    Tcl_PrintDouble(interp, (double) v[i].lv[2], c2);
    sprintf(buf + strlen(buf), "{%s %s %s}", c0, c1, c2);

    *freeProcPtr = (Tcl_FreeProc *) free;
    return buf;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectSurfacesParseProc:
 *
 * process the '-surfaces' option of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectSurfacesParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;
    Tcl_Interp *interp;
    Tk_Window tkwin;
    char *value;
    char *widgRec;
    int offset;
{
    int i, j, index, argc, nc;
    char **argv = NULL;
    char **comp = NULL;
    Surface *s = NULL;
    Object *obj = (Object *) widgRec;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }

    if (argc == 0) {
	if (obj->s) {
	    free((void *) obj->s);
	    obj->s = NULL;
	}
	obj->ns = 0;
	free((void *) argv);
	return TCL_OK;
    }

    if ((s = (Surface *) malloc(argc * sizeof(Surface))) == NULL) {
	Tcl_AppendResult(interp, "failed to allocate memory for surface list.", (char *) NULL);
	goto error;
    }

    for (i = 0; i < argc; i++) {

	if (Tcl_SplitList(interp, argv[i], &nc, &comp) != TCL_OK) {
	    goto error;
	}

	if (nc < 3) {
	    Tcl_AppendResult(interp, "each surface must contain at least one component.", (char *) NULL);
	    goto error;
	}

	if ((s[i].index = (int *) malloc((nc + 1) * sizeof(int))) == NULL) {
	    Tcl_AppendResult(interp, "failed to allocate memory for surface vertex indices.", (char *) NULL);
	    goto error;
	}

	if ((s[i].lvn = (Normal3D *) malloc((nc + 1) * sizeof(Normal3D))) == NULL) {
	    Tcl_AppendResult(interp, "failed to allocate memory for vertex normals.", (char *) NULL);
	    goto error;
	}

	for (j = 0; j < nc; j++) {
	    if (Tcl_GetInt(interp, comp[j], &index) != TCL_OK) {
		goto error;
	    }
	    if (index < 1) {
		Tcl_AppendResult(interp, "vertex index must be positive.", (char *) NULL);
		goto error;
	    }
	    s[i].index[j] = index - 1;
	}

	if (s[i].index[nc - 1] != s[i].index[0]) {
	    s[i].index[nc] = s[i].index[0];
	    nc++;
	}
	s[i].id = i + 1;
	s[i].vcnt = nc;
	s[i].obj = obj;
	s[i].normal = SM_VERTEX_NORMAL_DEFAULT;

	free((void *) comp);
	comp = NULL;
    }

    if (obj->s) free((void *) obj->s);
    free((void *) argv);
    obj->ns = argc;
    obj->s = s;
    obj->updateFlag |= SM_OBJECT_UPDATE_NORMAL + SM_OBJECT_UPDATE_POSITION;
    return TCL_OK;

  error:
    free((void *) argv);
    if (comp) free((void *) comp);
    if (s->index) free((void *) s->index);
    if (s->lvn) free((void *) s->lvn);
    if (s) free((void *) s);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectSurfacesPrintProc:
 *
 * produce a printable string for the '-surfaces' option.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectSurfacesPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    int i, j;
    int cnt = 0;
    char *buf;
    Surface *s;
    Object *obj = (Object *) widgRec;

    if (obj->ns == 0) {
	*freeProcPtr = NULL;
	return "";
    }

    for (s = obj->s, i = 0; i < obj->ns; i++, s++)
	cnt += s->vcnt;

    if ((buf = (char *) malloc(cnt * 12)) == NULL) {
	*freeProcPtr = NULL;
	return "";
    }

    *buf = '\0';
    for (s = obj->s, i = 0; i < obj->ns - 1; i++, s++) {
	strcat(buf, "{");
	for (j = 0; j < s->vcnt - 1; j++) {
	    sprintf(buf + strlen(buf), "%u ", s->index[j] + 1);
	}
	sprintf(buf + strlen(buf), "%u} ", s->index[j] + 1);
    }
    strcat(buf, "{");
    for (j = 0; j < s->vcnt - 1; j++) {
	sprintf(buf + strlen(buf), "%u ", s->index[j] + 1);
    }
    sprintf(buf + strlen(buf), "%u}", s->index[j] + 1);

    *freeProcPtr = (Tcl_FreeProc *) free;
    return buf;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectNormalsParseProc:
 *
 * process the '-normals' option of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectNormalsParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;
    Tcl_Interp *interp;
    Tk_Window tkwin;
    char *value;
    char *widgRec;
    int offset;
{
    Object *obj = (Object *) widgRec;

    /* we simply save a copy of the value string here; actual parsing is done elsewhere */
    if (obj->normals) {
	free((void *) obj->normals);
    }

    if ((*value)) {
	if ((obj->normals = (char *) malloc(strlen(value) + 1))) {
	    strcpy(obj->normals, value);
	}
    }
    else {
	obj->normals = NULL;
    }
    obj->updateFlag |= SM_OBJECT_UPDATE_NORMAL;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectNormalsPrintProc:
 *
 * produce a printable string for the '-normals' option.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectNormalsPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *obj = (Object *) widgRec;

#ifdef DEBUG
    int i, j;
    Normal3D *lvn;
    Vertex *v;
    Surface *s = obj->s;

    if (obj->s && obj->v) {
	for (i = 0, s = obj->s; i < obj->ns; i++, s++) {
	    if (s->normal) {
		printf("\nindex %d (specified):\n", s->id);
		for (j = 0, lvn = s->lvn; j < s->vcnt; j++, lvn++) {
		    printf("    {%f %f %f}\n", lvn->v[0], lvn->v[1], lvn->v[2]);
		}
		printf("\n");
	    }
	    else {
		printf("\nindex %d (averaged):\n", s->id);
		for (j = 0; j < s->vcnt; j++) {
		    v = &obj->v[s->index[j] - 1];
		    printf("    {%f %f %f}\n", v->lvn.v[0], v->lvn.v[1], v->lvn.v[2]);
		}
		printf("\n");

	    }
	}
    }
#endif

    *freeProcPtr = NULL;
    if (obj->normals) {
	return obj->normals;
    }
    else {
	return "";
    }
}

/*
 *--------------------------------------------------------------
 *
 * ObjectShadingParseProc:
 *
 * process the '-shading' option of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectShadingParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int i, len;
    char *p, c;
    Object *obj = (Object *) widgRec;

    c = *value;
    len = strlen(value);
    i = 0;
    p = surfaceShadingMethods[0];
    while (p) {
	if ((c == *p) && (strncmp(p, value, len) == 0) && (len >= 2)) {
	    obj->shading = i;
	    return TCL_OK;
	}
	p = surfaceShadingMethods[++i];
    }
    Tcl_AppendResult(interp, "bad shading method \"", value, "\": should be either flat or smooth.", (char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectShadingPrintProc:
 *
 * produce a printable string for the '-shading' option.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectShadingPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    *freeProcPtr = NULL;
    return surfaceShadingMethods[((Object *) widgRec)->shading];
}

/*
 *--------------------------------------------------------------
 *
 * ObjectCoefficientParseProc:
 *
 * process reflection coefficient options of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectCoefficientParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int argc;
    char **argv;
    double r, g, b;
    Object *obj = (Object *) widgRec;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc != 3) {
	Tcl_AppendResult(interp, "bad format: must specify three coefficients.", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }
    if (Tcl_GetDouble(interp, argv[0], &r) != TCL_OK ||
	Tcl_GetDouble(interp, argv[1], &g) != TCL_OK ||
	Tcl_GetDouble(interp, argv[2], &b) != TCL_OK) {
	free((void *) argv);
	return TCL_ERROR;
    }
    if ((r < 0) || (r > 1) || (g < 0) || (g > 1) || (b < 0) || (b > 1)) {
	Tcl_AppendResult(interp, "bad value: coefficients must be between 0 and 1.", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }

    free((void *) argv);
    switch((int) clientData) {

      case SM_OBJECT_AMBIENT:
	obj->ambient[0] = (DOUBLE) r;
	obj->ambient[1] = (DOUBLE) g;
	obj->ambient[2] = (DOUBLE) b;
	obj->updateFlag |= SM_OBJECT_UPDATE_AMBIENT;
	break;

      case SM_OBJECT_DIFFUSE:
	obj->diffuse[0] = (DOUBLE) r;
	obj->diffuse[1] = (DOUBLE) g;
	obj->diffuse[2] = (DOUBLE) b;
	obj->updateFlag |= SM_OBJECT_UPDATE_DIFFUSE;
	break;

      case SM_OBJECT_EMISSION:
	obj->emission[0] = (DOUBLE) r;
	obj->emission[1] = (DOUBLE) g;
	obj->emission[2] = (DOUBLE) b;
	obj->updateFlag |= SM_OBJECT_UPDATE_EMISSION;
	break;

      case SM_OBJECT_BF_AMBIENT:
	obj->bfambient[0] = (DOUBLE) r;
	obj->bfambient[1] = (DOUBLE) g;
	obj->bfambient[2] = (DOUBLE) b;
	obj->updateFlag |= SM_OBJECT_UPDATE_BF_AMBIENT;
	break;

      case SM_OBJECT_BF_DIFFUSE:
	obj->bfdiffuse[0] = (DOUBLE) r;
	obj->bfdiffuse[1] = (DOUBLE) g;
	obj->bfdiffuse[2] = (DOUBLE) b;
	obj->updateFlag |= SM_OBJECT_UPDATE_BF_DIFFUSE;
	break;

      case SM_OBJECT_BF_EMISSION:
	obj->bfemission[0] = (DOUBLE) r;
	obj->bfemission[1] = (DOUBLE) g;
	obj->bfemission[2] = (DOUBLE) b;
	obj->updateFlag |= SM_OBJECT_UPDATE_BF_EMISSION;
	break;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectCoefficientPrintPRoc:
 *
 * produce a printable string for reflection coefficients.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectCoefficientPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    static char buf[256];
    DOUBLE *v = NULL;
    Object *obj = (Object *) widgRec;
    char c0[TCL_DOUBLE_SPACE], c1[TCL_DOUBLE_SPACE];
    char c2[TCL_DOUBLE_SPACE];
    Tcl_Interp *interp = obj->cell->interp;

    switch((int) clientData) {
      case SM_OBJECT_AMBIENT:
	v = obj->ambient;
	break;

      case SM_OBJECT_DIFFUSE:
	v = obj->diffuse;
	break;

      case SM_OBJECT_EMISSION:
	v = obj->emission;
	break;

      case SM_OBJECT_BF_AMBIENT:
	v = obj->bfambient;
	break;

      case SM_OBJECT_BF_DIFFUSE:
	v = obj->bfdiffuse;
	break;

      case SM_OBJECT_BF_EMISSION:
	v = obj->bfemission;
	break;
    }
    Tcl_PrintDouble(interp, (double) v[0], c0);
    Tcl_PrintDouble(interp, (double) v[1], c1);
    Tcl_PrintDouble(interp, (double) v[2], c2);
    sprintf(buf, "%s %s %s", c0, c1, c2);
    *freeProcPtr = NULL;
    return buf;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectSpecularParseProc:
 *
 * process specular reflection options of an object.
 *
 *--------------------------------------------------------------
 */

static int
ObjectSpecularParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int argc;
    char **argv;
    double exponent;
    XColor *color = NULL;
    Object *obj = (Object *) widgRec;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc != 2) {
	Tcl_AppendResult(interp, "bad format: must specify {color exponent}.", (char *) NULL);
	goto err;
    }

    /* parse specular color */
    if ((color = Tk_GetColor(interp, tkwin, Tk_GetUid(argv[0]))) == NULL) {
	goto err;
    }
    if (Tcl_GetDouble(interp, argv[1], &exponent) != TCL_OK) {
	goto err;
    }
    if (exponent < 0) {
	Tcl_AppendResult(interp, "bad value: specular reflection exponent must be non-negative.", (char *) NULL);
	goto err;
    }
    if (clientData == (ClientData) SM_OBJECT_SPECULAR) {
	obj->specular.color = color;
	obj->specular.exponent = exponent;
	obj->updateFlag |= SM_OBJECT_UPDATE_SPECULAR;
    } else if (clientData == (ClientData) SM_OBJECT_BF_SPECULAR) {
	obj->bfspecular.color = color;
	obj->bfspecular.exponent = exponent;
	obj->updateFlag |= SM_OBJECT_UPDATE_BF_SPECULAR;
    }
    free((void *) argv);
    return TCL_OK;

  err:
    if (color) {Tk_FreeColor(color);}
    free((void *) argv);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ObjectSpeculartPrintPRoc:
 *
 * produce a printable string for specular reflection parameters.
 *
 *--------------------------------------------------------------
 */

static char *
ObjectSpecularPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    static char buf[256];
    Object *obj = (Object *) widgRec;
    char e0[TCL_DOUBLE_SPACE];
    Tcl_Interp *interp = obj->cell->interp;

    *freeProcPtr = NULL;
    if (clientData == (ClientData) SM_OBJECT_SPECULAR) {
	Tcl_PrintDouble(interp, (double) obj->specular.exponent, e0);
	sprintf(buf, "%s %s", Tk_NameOfColor(obj->specular.color), e0);
    } else {
	Tcl_PrintDouble(interp, (double) obj->bfspecular.exponent, e0);
	sprintf(buf, "%s %s", Tk_NameOfColor(obj->bfspecular.color), e0);
    }
    return buf;
}

/*
 *--------------------------------------------------------------
 *
 * PolygonModeParseProc:
 *
 * process the polygon drawing mode options (-frontmode, -backmode)
 * of an object.
 *
 *--------------------------------------------------------------
 */

static int
PolygonModeParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int len;
    GLenum *mode;
    Object *obj = (Object *) widgRec;

    len = strlen(value);
    mode = (((GLenum) clientData) == GL_FRONT) ? &obj->frontmode : &obj->backmode;
    if ((strncmp("point", value, len) == 0) && (len >= 2)) {
	*mode = GL_POINT;
	return TCL_OK;
    }
    if ((strncmp("line", value, len) == 0) && (len >= 2)) {
	*mode = GL_LINE;
	return TCL_OK;
    }
    if ((strncmp("fill", value, len) == 0) && (len >= 2)) {
	*mode = GL_FILL;
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad polygon drawing mode \"", value, "\": should be one of point, line, or fill.", (char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * PolygonModePrintProc:
 *
 * produce a printable string for the polygon drawing mode
 * (-frontmode, -backmode).
 *
 *--------------------------------------------------------------
 */

static char *
PolygonModePrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *obj = (Object *) widgRec;
    GLenum *mode;

    *freeProcPtr = NULL;
    mode = (((GLenum) clientData) == GL_FRONT) ? &obj->frontmode : &obj->backmode;
    if (*mode == GL_POINT) {
	return "point";
    }
    if (*mode == GL_LINE) {
	return "line";
    }
    if (*mode == GL_FILL) {
	return "fill";
    }
    return "unknown polygon drawing mode";
}

/*
 *--------------------------------------------------------------
 *
 * PolygonCullFaceParseProc:
 *
 * process the polygon face culling option -cull.
 *
 *--------------------------------------------------------------
 */

static int
PolygonCullFaceParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int len;
    Object *obj = (Object *) widgRec;

    len = strlen(value);

    if ((strncmp("none", value, len) == 0) && (len >= 2)) {
	obj->cull = 0;
	return TCL_OK;
    }
    if ((strncmp("front", value, len) == 0) && (len >= 2)) {
	obj->cull = GL_FRONT;
	return TCL_OK;
    }
    if ((strncmp("back", value, len) == 0) && (len >= 2)) {
	obj->cull = GL_BACK;
	return TCL_OK;
    }
    if ((strncmp("both", value, len) == 0) && (len >= 2)) {
	obj->cull = GL_FRONT_AND_BACK;
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad polygon face culling mode \"", value, "\": should be one of none, front, back, or both.", (char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * PolygonCullFacePrintProc:
 *
 * produce a printable string for the polygon culling mode
 * (-cull).
 *
 *--------------------------------------------------------------
 */

static char *
PolygonCullFacePrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *obj = (Object *) widgRec;

    *freeProcPtr = NULL;
    if (obj->cull == 0) {
	return "none";
    }
    if (obj->cull == GL_FRONT) {
	return "front";
    }
    if (obj->cull == GL_BACK) {
	return "back";
    }
    if (obj->cull == GL_FRONT_AND_BACK) {
	return "both";
    }
    return "unknown polygon culling mode";
}

/*
 *--------------------------------------------------------------
 *
 * QuadricDrawStyleParseProc:
 *
 * process the drawing style options (-drawstyle)
 * of a quadric object.
 *
 *--------------------------------------------------------------
 */

static int
QuadricDrawStyleParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int len;
    Object *obj = (Object *) widgRec;

    len = strlen(value);
    if ((strncmp("point", value, len) == 0) && (len >= 2)) {
	obj->drawStyle = GLU_POINT;
	return TCL_OK;
    }
    if ((strncmp("line", value, len) == 0) && (len >= 2)) {
	obj->drawStyle = GLU_LINE;
	return TCL_OK;
    }
    if ((strncmp("fill", value, len) == 0) && (len >= 2)) {
	obj->drawStyle = GLU_FILL;
	return TCL_OK;
    }
    if ((strncmp("silhouette", value, len) == 0) && (len >= 2)) {
	obj->drawStyle = GLU_SILHOUETTE;
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad quadric drawing style \"", value,
		     "\": should be one of point, line, silhouette, or fill.", (char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * QuadricDrawStylePrintProc:
 *
 * produce a printable string for the quadric drawing style
 * (-drawstyle).
 *
 *--------------------------------------------------------------
 */

static char *
QuadricDrawStylePrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *obj = (Object *) widgRec;

    *freeProcPtr = NULL;
    if (obj->drawStyle == GLU_POINT) {
	return "point";
    }
    if (obj->drawStyle == GLU_LINE) {
	return "line";
    }
    if (obj->drawStyle == GLU_FILL) {
	return "fill";
    }
    if (obj->drawStyle == GLU_SILHOUETTE) {
	return "silhouette";
    }
    return "unknown quadric drawing style";
}

/*
 *--------------------------------------------------------------
 *
 * QuadricNormalsParseProc:
 *
 * process the normals options (-normals)
 * of a quadric object.
 *
 *--------------------------------------------------------------
 */

static int
QuadricNormalsParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int len;
    Object *obj = (Object *) widgRec;

    len = strlen(value);
    if ((strncmp("none", value, len) == 0) && (len >= 2)) {
	obj->quadNormals = GLU_NONE;
	return TCL_OK;
    }
    if ((strncmp("flat", value, len) == 0) && (len >= 2)) {
	obj->quadNormals = GLU_FLAT;
	return TCL_OK;
    }
    if ((strncmp("smooth", value, len) == 0) && (len >= 2)) {
	obj->quadNormals = GLU_SMOOTH;
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad quadric normals \"", value,
		     "\": should be one of none, flat, or smooth.", (char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * QuadricNormalsPrintProc:
 *
 * produce a printable string for the quadric drawing style
 * (-normals).
 *
 *--------------------------------------------------------------
 */

static char *
QuadricNormalsPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *obj = (Object *) widgRec;

    *freeProcPtr = NULL;
    if (obj->quadNormals == GLU_NONE) {
	return "none";
    }
    if (obj->quadNormals == GLU_FLAT) {
	return "flat";
    }
    if (obj->quadNormals == GLU_SMOOTH) {
	return "smooth";
    }
    return "unknown quadric normals";
}

/*
 *--------------------------------------------------------------
 *
 * QuadricOrientationParseProc:
 *
 * process the orientation options (-orientation)
 * of a quadric object.
 *
 *--------------------------------------------------------------
 */

static int
QuadricOrientationParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int len;
    Object *obj = (Object *) widgRec;

    len = strlen(value);
    if ((strncmp("outside", value, len) == 0) && (len >= 2)) {
	obj->orientation = GLU_OUTSIDE;
	return TCL_OK;
    }
    if ((strncmp("inside", value, len) == 0) && (len >= 2)) {
	obj->orientation = GLU_INSIDE;
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad quadric orientation \"", value,
		     "\": should be either inside or outside.", (char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * QuadricOrientationPrintProc:
 *
 * produce a printable string for the quadric drawing style
 * (-orientation).
 *
 *--------------------------------------------------------------
 */

static char *
QuadricOrientationPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    Object *obj = (Object *) widgRec;

    *freeProcPtr = NULL;
    if (obj->orientation == GLU_OUTSIDE) {
	return "outside";
    }
    if (obj->orientation == GLU_INSIDE) {
	return "inside";
    }
    return "unknown quadric orientation";
}
