/* 
 * smvoption.c --
 *
 *	This file implements canvas viewport options.
 *
 */

#include "copyright.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <tk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "smInt.h"


/*
 * Prototypes for procedures defined externally
 */

Cell *		Sm_FindCell _ANSI_ARGS_((Tcl_Interp *interp, char *label));

void		SmInsertPortIntoCell _ANSI_ARGS_((Cell *cell, PortItem *port));

void		SmRemovePortFromCell _ANSI_ARGS_((Cell *cell, PortItem *port));

void		SmFastNormalizeVector3D _ANSI_ARGS_ ((DOUBLE vin[], DOUBLE vout[]));

void		SmDisplayObjects _ANSI_ARGS_ ((PortItem *port, int pick));

void		SmSelectViewport _ANSI_ARGS_ ((PortItem *port));

void            SmViewportUpdate _ANSI_ARGS_((ClientData clientData));


/*
 * Prototypes for procedures defined in this file that are exported:
 */

int		SmVectorParseProc _ANSI_ARGS_((ClientData clientData,
						Tcl_Interp *interp, Tk_Window tkwin,
						char *value, char *widgRec, int offset));

char *		SmVectorPrintProc _ANSI_ARGS_((ClientData clientData,
						Tk_Window tkwin, char *widgRec, int offset,
						Tcl_FreeProc **freeProcPtr));

void		SmSetClientData _ANSI_ARGS_((ClientData clientData));

int		SmViewportChanged _ANSI_ARGS_((PortItem *port));


/*
 * Prototypes for procedures defined in this file that are static:
 */

static int	PortProjectionParseProc _ANSI_ARGS_((ClientData clientData,
						     Tcl_Interp *interp, Tk_Window tkwin,
						     char *value, char *widgRec, int offset));

static char *	PortProjectionPrintProc _ANSI_ARGS_((ClientData clientData,
						     Tk_Window tkwin, char *widgRec, int offset,
						     Tcl_FreeProc **freeProcPtr));

static int	PortCellParseProc _ANSI_ARGS_((ClientData clientData,
					       Tcl_Interp *interp, Tk_Window tkwin,
					       char *value, char *widgRec, int offset));

static char *	PortCellPrintProc _ANSI_ARGS_((ClientData clientData,
					       Tk_Window tkwin, char *widgRec, int offset,
					       Tcl_FreeProc **freeProcPtr));

static int	PortObjectVisibleMaskParseProc _ANSI_ARGS_((ClientData clientData,
							    Tcl_Interp *interp, Tk_Window tkwin,
							    char *value, char *widgRec, int offset));

static char *	PortObjectVisibleMaskPrintProc _ANSI_ARGS_((ClientData clientData,
							    Tk_Window tkwin, char *widgRec, int offset,
							    Tcl_FreeProc **freeProcPtr));

static int	PortObjectPickParseProc _ANSI_ARGS_((ClientData clientData,
						     Tcl_Interp *interp, Tk_Window tkwin,
						     char *value, char *widgRec, int offset));

static char *	PortObjectPickPrintProc _ANSI_ARGS_((ClientData clientData,
						     Tk_Window tkwin, char *widgRec, int offset,
						     Tcl_FreeProc **freeProcPtr));

static int	PortLightSourceParseProc _ANSI_ARGS_((ClientData clientData,
						      Tcl_Interp *interp, Tk_Window tkwin,
						      char *value, char *widgRec, int offset));

static char *	PortLightSourcePrintProc _ANSI_ARGS_((ClientData clientData,
						      Tk_Window tkwin, char *widgRec, int offset,
						      Tcl_FreeProc **freeProcPtr));

static int	PortDepthCueParseProc _ANSI_ARGS_((ClientData clientData,
						   Tcl_Interp *interp, Tk_Window tkwin,
						   char *value, char *widgRec, int offset));

static char *	PortDepthCuePrintProc _ANSI_ARGS_((ClientData clientData,
						   Tk_Window tkwin, char *widgRec, int offset,
						   Tcl_FreeProc **freeProcPtr));


static int	ParseLightDirection _ANSI_ARGS_((Tcl_Interp *interp, char *value, float dir[3], float ndir[3]));

static int	ParseLightPosition _ANSI_ARGS_((Tcl_Interp *interp, char *value, float pos[4]));

static int	ParseLightAttenuation _ANSI_ARGS_((Tcl_Interp *interp, char *value, float att[3]));


/*
 * Custom options for viewport item type
 */

static Tk_CustomOption tk_ViewportOptionProjection = {
    PortProjectionParseProc,
    PortProjectionPrintProc,
    (ClientData) NULL,
};

static Tk_CustomOption tk_ViewportOptionVRP = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_PORT_VRP,
};

static Tk_CustomOption tk_ViewportOptionVPN = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_PORT_VPN,
};

static Tk_CustomOption tk_ViewportOptionVUP = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_PORT_VUP,
};

static Tk_CustomOption tk_ViewportOptionPRP = {
    SmVectorParseProc,
    SmVectorPrintProc,
    (ClientData) SM_PORT_PRP,
};

static Tk_CustomOption tk_ViewportOptionLightSource0 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_0,
};

static Tk_CustomOption tk_ViewportOptionLightSource1 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_1,
};

static Tk_CustomOption tk_ViewportOptionLightSource2 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_2,
};

static Tk_CustomOption tk_ViewportOptionLightSource3 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_3,
};

static Tk_CustomOption tk_ViewportOptionLightSource4 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_4,
};

static Tk_CustomOption tk_ViewportOptionLightSource5 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_5,
};

static Tk_CustomOption tk_ViewportOptionLightSource6 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_6,
};

static Tk_CustomOption tk_ViewportOptionLightSource7 = {
    PortLightSourceParseProc,
    PortLightSourcePrintProc,
    (ClientData) SM_LIGHT_SOURCE_7,
};

static Tk_CustomOption tk_ViewportOptionCell = {
    PortCellParseProc,
    PortCellPrintProc,
    (ClientData) NULL,
};

static Tk_CustomOption tk_ViewportOptionVisibleMask = {
    PortObjectVisibleMaskParseProc,
    PortObjectVisibleMaskPrintProc,
    (ClientData) NULL,
};

static Tk_CustomOption tk_ViewportOptionPick = {
    PortObjectPickParseProc,
    PortObjectPickPrintProc,
    (ClientData) SM_PORT_PICK,
};

static Tk_CustomOption tk_ViewportOptionDepthCue = {
    PortDepthCueParseProc,
    PortDepthCuePrintProc,
    (ClientData) NULL,
};

/*
 * Information used for parsing canvas viewport configuration specs:
 */

Tk_ConfigSpec canvasConfigSpecs[] = {

    {TK_CONFIG_COLOR, "-ambientlight", "ambientLight", "AmbientLight",
       "black", Tk_Offset(PortItem, ambientLight), 0},

    {TK_CONFIG_DOUBLE, "-bp", "backplane", "Backplane",
       "-1e6", Tk_Offset(PortItem, bp), 0},

    {TK_CONFIG_CUSTOM, "-cell", "cell", "Cell",
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionCell},

    {TK_CONFIG_CUSTOM, "-depthcue", "depthCue", "DepthCue",
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionDepthCue},

    {TK_CONFIG_DOUBLE, "-epsilon", "epsilon", "Epsilon",
       "1e-3", Tk_Offset(PortItem, epsilon), 0},

    {TK_CONFIG_COLOR, "-fill", "background", "Background",
       "black", Tk_Offset(PortItem, fillColor), 0},

    {TK_CONFIG_BOOLEAN, "-hsr", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(PortItem, hsr), 0},

    {TK_CONFIG_CUSTOM, "-light0", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource0},

    {TK_CONFIG_CUSTOM, "-light1", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource1},

    {TK_CONFIG_CUSTOM, "-light2", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource2},

    {TK_CONFIG_CUSTOM, "-light3", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource3},

    {TK_CONFIG_CUSTOM, "-light4", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource4},

    {TK_CONFIG_CUSTOM, "-light5", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource5},

    {TK_CONFIG_CUSTOM, "-light6", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource6},

    {TK_CONFIG_CUSTOM, "-light7", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource7},

    {TK_CONFIG_DOUBLE, "-maxu", (char *) NULL, (char *) NULL,
       "100.0", Tk_Offset(PortItem, maxu), 0},

    {TK_CONFIG_DOUBLE, "-maxv", (char *) NULL, (char *) NULL,
       "100.0", Tk_Offset(PortItem, maxv), 0},

    {TK_CONFIG_DOUBLE, "-minu", (char *) NULL, (char *) NULL,
       "-100.0", Tk_Offset(PortItem, minu), 0},

    {TK_CONFIG_DOUBLE, "-minv", (char *) NULL, (char *) NULL,
       "-100.0", Tk_Offset(PortItem, minv), 0},

    {TK_CONFIG_COLOR, "-outline", "foreground", "Foreground",
       "black", Tk_Offset(PortItem, outlineColor), TK_CONFIG_NULL_OK},

    {TK_CONFIG_CUSTOM, "-pick", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionPick},

    {TK_CONFIG_CUSTOM, "-projection", (char *) NULL, (char *) NULL,
       "parallel", 0, 0, &tk_ViewportOptionProjection},

    {TK_CONFIG_CUSTOM, "-prp", (char *) NULL, (char *) NULL,
       "0.0 0.0 100.0", 0, 0, &tk_ViewportOptionPRP},

    {TK_CONFIG_CUSTOM, "-tags", (char *) NULL, (char *) NULL,
       (char *) NULL, 0, TK_CONFIG_NULL_OK, &tk_CanvasTagsOption},

    {TK_CONFIG_SYNONYM, "-vismask", "visibleMask", (char *) NULL,
       (char *) NULL, 0, 0, NULL},

    {TK_CONFIG_CUSTOM, "-visiblemask", "visibleMask", "VisibleMask",
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionVisibleMask},

    {TK_CONFIG_CUSTOM, "-vpn", (char *) NULL, (char *) NULL,
       "0.0 0.0 1.0", 0, 0, &tk_ViewportOptionVPN},

    {TK_CONFIG_CUSTOM, "-vrp", (char *) NULL, (char *) NULL,
       "0.0 0.0 0.0",  0, 0, &tk_ViewportOptionVRP},

    {TK_CONFIG_CUSTOM, "-vup", (char *) NULL, (char *) NULL,
       "0.0 1.0 0.0", 0, 0, &tk_ViewportOptionVUP},

    {TK_CONFIG_PIXELS, "-width", (char *) NULL, (char *) NULL,
       "1", Tk_Offset(PortItem, width), 0, NULL},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Information used for parsing viewport configuration specs:
 */

Tk_ConfigSpec widgetConfigSpecs[] = {

    {TK_CONFIG_COLOR, "-ambientlight", "ambientLight", "AmbientLight",
       "black", Tk_Offset(PortItem, ambientLight), 0},

    {TK_CONFIG_DOUBLE, "-bp", "backplane", "Backplane",
       "-1e6", Tk_Offset(PortItem, bp), 0},

    {TK_CONFIG_CUSTOM, "-cell", "cell", "Cell",
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionCell},

    {TK_CONFIG_CUSTOM, "-depthcue", "depthCue", "DepthCue",
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionDepthCue},

    {TK_CONFIG_DOUBLE, "-epsilon", "epsilon", "Epsilon",
       "1e-3", Tk_Offset(PortItem, epsilon), 0},

    {TK_CONFIG_COLOR, "-fill", "background", "Background",
       "black", Tk_Offset(PortItem, fillColor), 0},

    {TK_CONFIG_BOOLEAN, "-hsr", (char *) NULL, (char *) NULL,
       "on", Tk_Offset(PortItem, hsr), 0},

    {TK_CONFIG_CUSTOM, "-light0", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource0},

    {TK_CONFIG_CUSTOM, "-light1", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource1},

    {TK_CONFIG_CUSTOM, "-light2", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource2},

    {TK_CONFIG_CUSTOM, "-light3", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource3},

    {TK_CONFIG_CUSTOM, "-light4", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource4},

    {TK_CONFIG_CUSTOM, "-light5", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource5},

    {TK_CONFIG_CUSTOM, "-light6", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource6},

    {TK_CONFIG_CUSTOM, "-light7", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionLightSource7},

    {TK_CONFIG_DOUBLE, "-maxu", (char *) NULL, (char *) NULL,
       "100.0", Tk_Offset(PortItem, maxu), 0},

    {TK_CONFIG_DOUBLE, "-maxv", (char *) NULL, (char *) NULL,
       "100.0", Tk_Offset(PortItem, maxv), 0},

    {TK_CONFIG_DOUBLE, "-minu", (char *) NULL, (char *) NULL,
       "-100.0", Tk_Offset(PortItem, minu), 0},

    {TK_CONFIG_DOUBLE, "-minv", (char *) NULL, (char *) NULL,
       "-100.0", Tk_Offset(PortItem, minv), 0},

    {TK_CONFIG_COLOR, "-outline", "foreground", "Foreground",
       "black", Tk_Offset(PortItem, outlineColor), TK_CONFIG_NULL_OK},

    {TK_CONFIG_CUSTOM, "-pick", (char *) NULL, (char *) NULL,
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionPick},

    {TK_CONFIG_CUSTOM, "-projection", (char *) NULL, (char *) NULL,
       "parallel", 0, 0, &tk_ViewportOptionProjection},

    {TK_CONFIG_CUSTOM, "-prp", (char *) NULL, (char *) NULL,
       "0.0 0.0 100.0", 0, 0, &tk_ViewportOptionPRP},

    {TK_CONFIG_SYNONYM, "-vismask", "visibleMask", (char *) NULL,
       (char *) NULL, 0, 0, NULL},

    {TK_CONFIG_CUSTOM, "-visiblemask", "visibleMask", "VisibleMask",
       "", 0, TK_CONFIG_NULL_OK, &tk_ViewportOptionVisibleMask},

    {TK_CONFIG_CUSTOM, "-vpn", (char *) NULL, (char *) NULL,
       "0.0 0.0 1.0", 0, 0, &tk_ViewportOptionVPN},

    {TK_CONFIG_CUSTOM, "-vrp", (char *) NULL, (char *) NULL,
       "0.0 0.0 0.0",  0, 0, &tk_ViewportOptionVRP},

    {TK_CONFIG_CUSTOM, "-vup", (char *) NULL, (char *) NULL,
       "0.0 1.0 0.0", 0, 0, &tk_ViewportOptionVUP},

    {TK_CONFIG_PIXELS, "-width", (char *) NULL, (char *) NULL,
       "1", Tk_Offset(PortItem, width), 0, NULL},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};



/*
 *--------------------------------------------------------------
 *
 * SmViewportChanged:
 *
 * determine if viewport specification has changed.
 *
 *--------------------------------------------------------------
 */

int
SmViewportChanged(port)
     PortItem *port;
{
    if (port->canvas) {
	return ((canvasConfigSpecs[1].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* bp */
		(canvasConfigSpecs[4].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* epsilon */
		(canvasConfigSpecs[15].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* maxu */
		(canvasConfigSpecs[16].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* maxv */
		(canvasConfigSpecs[17].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* minu */
		(canvasConfigSpecs[18].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* minv */
		(canvasConfigSpecs[21].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* projection */
		(canvasConfigSpecs[22].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* prp */
		(canvasConfigSpecs[26].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* vpn */
		(canvasConfigSpecs[27].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* vrp */
		(canvasConfigSpecs[28].specFlags & TK_CONFIG_OPTION_SPECIFIED)		/* vup */
		) ? 1 : 0;
    } else {
	return ((widgetConfigSpecs[1].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* bp */
		(widgetConfigSpecs[4].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* epsilon */
		(widgetConfigSpecs[15].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* maxu */
		(widgetConfigSpecs[16].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* maxv */
		(widgetConfigSpecs[17].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* minu */
		(widgetConfigSpecs[18].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* minv */
		(widgetConfigSpecs[21].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* projection */
		(widgetConfigSpecs[22].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* prp */
		(widgetConfigSpecs[25].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* vpn */
		(widgetConfigSpecs[26].specFlags & TK_CONFIG_OPTION_SPECIFIED) ||	/* vrp */
		(widgetConfigSpecs[27].specFlags & TK_CONFIG_OPTION_SPECIFIED)		/* vup */
		) ? 1 : 0;
    }
}

/*
 *--------------------------------------------------------------
 *
 * SmSetClientData:
 *
 * set clientData to point to the appropriate cell header
 *
 *--------------------------------------------------------------
 */

void
SmSetClientData(clientData)
     ClientData clientData;
{
    tk_ViewportOptionCell.clientData = clientData;
}

/*
 *--------------------------------------------------------------
 *
 * SmVectorParseProc:
 *
 * process options for items that take vectors.
 *
 *--------------------------------------------------------------
 */

int
SmVectorParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;              /* An index that indicates which option *
					 * is being parsed */
    Tcl_Interp *interp;                 /* Used for reporting errors. */
    Tk_Window tkwin;                    /* Window containing canvas widget. */
    char *value;                        /* Value of option (3 doubles). */
    char *widgRec;                      /* Pointer to record for item. */
    int offset;                         /* Offset into item (ignored). */
{
    register PortItem *port = (PortItem *) widgRec;
    register Object *obj = (Object *) widgRec;
    int argc;
    char **argv;
    double x, y, z;

    /* split the option value into components */

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
        return TCL_ERROR;
    }

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: 3 components required", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }

    if (Tcl_GetDouble(interp, argv[0], &x) != TCL_OK ||
	Tcl_GetDouble(interp, argv[1], &y) != TCL_OK ||
	Tcl_GetDouble(interp, argv[2], &z) != TCL_OK) {
	free((void *) argv);
	return TCL_ERROR;
    }
    free((void *) argv);
    switch((int) clientData) {

      case SM_PORT_VRP:
	port->vrp[0] = (DOUBLE) x;
	port->vrp[1] = (DOUBLE) y;
	port->vrp[2] = (DOUBLE) z;
	break;

      case SM_PORT_VPN:
	port->vpn[0] = (DOUBLE) x;
	port->vpn[1] = (DOUBLE) y;
	port->vpn[2] = (DOUBLE) z;
	break;
	
      case SM_PORT_VUP:
	port->vup[0] = (DOUBLE) x;
	port->vup[1] = (DOUBLE) y;
	port->vup[2] = (DOUBLE) z;
	break;
	
      case SM_PORT_PRP:
	port->prp[0] = (DOUBLE) x;
	port->prp[1] = (DOUBLE) y;
	port->prp[2] = (DOUBLE) z;
	break;

      case SM_OBJECT_FWD:
	obj->fwd[0] = (DOUBLE) x;
	obj->fwd[1] = (DOUBLE) y;
	obj->fwd[2] = (DOUBLE) z;
	obj->updateFlag |= SM_OBJECT_UPDATE_POSITION;
	break;

      case SM_OBJECT_UP:
	obj->up[0] = (DOUBLE) x;
	obj->up[1] = (DOUBLE) y;
	obj->up[2] = (DOUBLE) z;
	obj->updateFlag |= SM_OBJECT_UPDATE_POSITION;
	break;

      case SM_OBJECT_POS:
	obj->pos[0] = (DOUBLE) x;
	obj->pos[1] = (DOUBLE) y;
	obj->pos[2] = (DOUBLE) z;
	obj->updateFlag |= SM_OBJECT_UPDATE_POSITION;
	break;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SmVectorPrintProc:
 *
 * produce a printable string for those options for items
 * that take vector values.
 *
 *--------------------------------------------------------------
 */

char *
SmVectorPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
    ClientData clientData;              /* An index that indicates which option *
					 * is being parsed */
    Tk_Window tkwin;                    /* Window containing canvas widget. */
    char *widgRec;                      /* Pointer to record for item. */
    int offset;                         /* Ignored. */
    Tcl_FreeProc **freeProcPtr;         /* Pointer to variable to fill in with
                                         * information about how to reclaim
                                         * storage for return string. */
{
    static char buf[128];
    DOUBLE *arrayPtr = NULL;
    Tcl_Interp *interp = NULL;

    char c0[TCL_DOUBLE_SPACE], c1[TCL_DOUBLE_SPACE];
    char c2[TCL_DOUBLE_SPACE];

#if defined(__alpha__)
    switch((long) clientData) {
#else
    switch((int) clientData) {
#endif

      case SM_PORT_VRP:
	arrayPtr = ((PortItem *) widgRec)->vrp;
	interp = ((PortItem *) widgRec)->interp;
	break;

      case SM_PORT_VPN:
	arrayPtr = ((PortItem *) widgRec)->vpn;
	interp = ((PortItem *) widgRec)->interp;
	break;
	
      case SM_PORT_VUP:
	arrayPtr = ((PortItem *) widgRec)->vup;
	interp = ((PortItem *) widgRec)->interp;
	break;
	
      case SM_PORT_PRP:
	arrayPtr = ((PortItem *) widgRec)->prp;
	interp = ((PortItem *) widgRec)->interp;
	break;

      case SM_OBJECT_FWD:
	arrayPtr = ((Object *) widgRec)->fwd;
	interp = ((Object *) widgRec)->cell->interp;
	break;

      case SM_OBJECT_UP:
	arrayPtr = ((Object *) widgRec)->up;
	interp = ((Object *) widgRec)->cell->interp;
	break;

      case SM_OBJECT_POS:
	arrayPtr = ((Object *) widgRec)->pos;
	interp = ((Object *) widgRec)->cell->interp;
	break;

    }
    Tcl_PrintDouble(interp, (double) arrayPtr[0], c0);
    Tcl_PrintDouble(interp, (double) arrayPtr[1], c1);
    Tcl_PrintDouble(interp, (double) arrayPtr[2], c2);
    sprintf(buf, "%s %s %s", c0, c1, c2);
    *freeProcPtr = NULL;
    return buf;
}

/*
 *--------------------------------------------------------------
 *
 * PortProjectionParseProc:
 *
 * process '-projection' option for viewport items.
 *
 *--------------------------------------------------------------
 */

static int
PortProjectionParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;              /* not used */
    Tcl_Interp *interp;                 /* Used for reporting errors. */
    Tk_Window tkwin;                    /* Window containing canvas widget. */
    char *value;                        /* Value of option (parallel or perspective). */
    char *widgRec;                      /* Pointer to record for item. */
    int offset;                         /* Offset into item (ignored). */
{
    register PortItem *port = (PortItem *) widgRec;
    char c;
    int len, strncasecmp();

    c = *value;
    len = strlen(value);

    if ((c == 'p') && (len > 2) && (strncasecmp(value, "parallel", len) == 0)) {
	port->projection = SM_PORT_PARALLEL;
    } else if ((c == 'p') && (len > 2) && (strncasecmp(value, "perspective", len) == 0)) {
	port->projection = SM_PORT_PERSPECTIVE;
    } else {
	Tcl_AppendResult(interp, "illegal projection type: ", value,
			 ". should be one of: parallel, perspective", (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PortProjectionPrintProc:
 *
 * produce a printable string for the '-projection' option.
 *
 *--------------------------------------------------------------
 */

static char *
PortProjectionPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
    ClientData clientData;              /* not used */
    Tk_Window tkwin;                    /* Window containing canvas widget. */
    char *widgRec;                      /* Pointer to record for item. */
    int offset;                         /* Ignored. */
    Tcl_FreeProc **freeProcPtr;         /* Pointer to variable to fill in with
                                         * information about how to reclaim
                                         * storage for return string. */
{
    register PortItem *port = (PortItem *) widgRec;

    *freeProcPtr = NULL;
    switch(port->projection) {
      case SM_PORT_PARALLEL:
	return "parallel";

      case SM_PORT_PERSPECTIVE:
	return "perspective";

      default:
	return "unknown";
    }
}

/*
 *--------------------------------------------------------------
 *
 * PortCellParseProc:
 *
 * process '-cell' option for viewport items.
 *
 *--------------------------------------------------------------
 */

static int
PortCellParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;              /* not used */
    Tcl_Interp *interp;                 /* Used for reporting errors. */
    Tk_Window tkwin;                    /* Window containing canvas widget. */
    char *value;                        /* Value of option (parallel or perspective). */
    char *widgRec;                      /* Pointer to record for item. */
    int offset;                         /* Offset into item (ignored). */
{
    register PortItem *port = (PortItem *) widgRec;
    Cell *cell;

    if (!(*value)) {
	if (port->cell) {
	    SmRemovePortFromCell(port->cell, port);
	    port->cell = NULL;
	}
    }
    else if ((cell = Sm_FindCell(interp, value)) == NULL) {
	Tcl_AppendResult(interp, "cell \"", value, "\" does not exist.", (char *) NULL);
	return TCL_ERROR;
    }
    else {
	if (port->cell) {
	    SmRemovePortFromCell(port->cell, port);
	}
	port->cell = cell;
	SmInsertPortIntoCell(cell, port);
    }
    if (port->canvas) {
	port->redraw = 1;
    	Tk_CanvasEventuallyRedraw(port->canvas, ((Tk_Item *) port)->x1, ((Tk_Item *) port)->y1,
			          ((Tk_Item *) port)->x2, ((Tk_Item *) port)->y2);
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PortCellPrintProc:
 *
 * produce a printable string for the '-cell' option.
 *
 *--------------------------------------------------------------
 */

static char *
PortCellPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;              /* not used */
     Tk_Window tkwin;                    /* Window containing canvas widget. */
     char *widgRec;                      /* Pointer to record for item. */
     int offset;                         /* Ignored. */
     Tcl_FreeProc **freeProcPtr;         /* Pointer to variable to fill in with
					  * information about how to reclaim
					  * storage for return string. */
{
    register PortItem *port = (PortItem *) widgRec;

    *freeProcPtr = NULL;

    if (port->cell) {
	return port->cell->label;
    } else {
	return "";
    }
}

/*
 *--------------------------------------------------------------
 *
 * PortObjectVisibleMaskParseProc:
 *
 * process '-visibleMask' option for a viewport item.
 *
 *--------------------------------------------------------------
 */

static int
PortObjectVisibleMaskParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;              /* not used */
    Tcl_Interp *interp;                 /* Used for reporting errors. */
    Tk_Window tkwin;                    /* Window containing canvas widget. */
    char *value;                        /* Value of option (list of strings) */
    char *widgRec;                      /* Pointer to record for item. */
    int offset;                         /* Offset into item (ignored). */
{
    int i, argc, new;
    int *ptr;
    char **argv;
    Tcl_HashTable *hash;

    int inverted = 0;
    register PortItem *port = (PortItem *) widgRec;

    hash = &port->visibleHash;
    ptr = &port->invertedVisibleMask;

    Tcl_DeleteHashTable(hash);
    Tcl_InitHashTable(hash, TCL_STRING_KEYS);
    if (*value) {
	if (*value == '^') {
	    inverted = 1;
	    if (Tcl_SplitList(interp, value+1, &argc, &argv) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
	else if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	    return TCL_ERROR;
	}

	for (i = 0; i < argc; i++) {
	    (void) Tcl_CreateHashEntry(hash, argv[i], &new);
	}
	free((void *) argv);
    }
    if (port->visibleMaskStr) free((void *) port->visibleMaskStr);
    if ((port->visibleMaskStr = (char *) malloc(strlen(value) + 1))) {
	strcpy(port->visibleMaskStr, value);
    }
    *ptr = inverted;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PortObjectVisibleMaskPrintProc:
 *
 * produce a printable string for object visible mask options.
 *
 *--------------------------------------------------------------
 */

static char *
PortObjectVisibleMaskPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;              /* not used */
     Tk_Window tkwin;                    /* Window containing canvas widget. */
     char *widgRec;                      /* Pointer to record for item. */
     int offset;                         /* Ignored. */
     Tcl_FreeProc **freeProcPtr;         /* Pointer to variable to fill in with
					  * information about how to reclaim
					  * storage for return string. */
{
    register PortItem *port = (PortItem *) widgRec;

    *freeProcPtr = NULL;
    return port->visibleMaskStr;
}

/*
 *--------------------------------------------------------------
 *
 * PortObjectPickParseProc:
 *
 * process '-pick' option for a viewport item.
 *
 * syntax: {x y w h}
 *
 *--------------------------------------------------------------
 */

static int
PortObjectPickParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    PortItem *port = (PortItem *) widgRec;

    int argc;
    char **argv;

    if (!*value) {
	port->pick[0] = 0;
	return TCL_OK;
    } else {
	if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be {x y w h}", (char *) NULL);
	    free((void *) argv);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[0], &port->pick[1]) != TCL_OK ||
	    Tcl_GetInt(interp, argv[1], &port->pick[2]) != TCL_OK ||
	    Tcl_GetInt(interp, argv[2], &port->pick[3]) != TCL_OK ||
	    Tcl_GetInt(interp, argv[3], &port->pick[4]) != TCL_OK) {
	    free((void *) argv);
	    return TCL_ERROR;
	}
	free((void *) argv);
	port->pick[0] = 1;
	return TCL_OK;
    }
}

/*
 *--------------------------------------------------------------
 *
 * PortObjectPickPrintProc:
 *
 * produce the result of the pick operations.
 *
 *--------------------------------------------------------------
 */

static char *
PortObjectPickPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;              /* not used */
     Tk_Window tkwin;                    /* Window containing canvas widget. */
     char *widgRec;                      /* Pointer to record for item. */
     int offset;                         /* Ignored. */
     Tcl_FreeProc **freeProcPtr;         /* Pointer to variable to fill in with
					  * information about how to reclaim
					  * storage for return string. */
{
    static char buf[512];

    int i;
    int x, y;
    short cx, cy;
    GLint nhits, viewport[4];
    GLuint *hit, *front, depth, selectBuf[1024];
    Object *obj;
    DOUBLE msh[16] = {
	1, 0, 0, 0,
	0, 1, 0, 0,
	0, 0, 1, 0,
	0, 0, 0, 1,
    };


    PortItem *port = (PortItem *) widgRec;
    int *pick = port->pick;

    *freeProcPtr = NULL;
    if (port->pick[0]) {

	SmSelectViewport(port);
	glGetIntegerv(GL_VIEWPORT, viewport);

	if (port->canvas) {
	    Tk_CanvasWindowCoords(port->canvas, (double) 0, (double) 0, &cx, &cy);

	    x = pick[1] - port->bbox[0] - cx;
	    y = viewport[3] - (pick[2] - port->bbox[1] - cy);
	} else {
	    x = pick[1];
	    y = viewport[3] - pick[2];
	}

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPickMatrix((GLdouble) x,
		      (GLdouble) y,
		      (GLdouble) pick[3],
		      (GLdouble) pick[4],
		      viewport);

	switch(port->projection) {
	  case SM_PORT_PARALLEL:
	    msh[8] = -port->prp[0] / port->prp[2];
	    msh[9] = -port->prp[1] / port->prp[2];
	    glOrtho(port->minu - port->prp[0],
		    port->maxu - port->prp[0],
		    port->minv - port->prp[1],
		    port->maxv - port->prp[1],
		    port->prp[2], port->prp[2] - port->bp);
#ifdef USE_DOUBLE
	    glMultMatrixd(msh);
#else
	    glMultMatrixf(msh);
#endif
	    break;
	    
	  case SM_PORT_PERSPECTIVE:
	    glFrustum(port->minu - port->prp[0],
		      port->maxu - port->prp[0],
		      port->minv - port->prp[1],
		  port->maxv - port->prp[1],
		      port->prp[2], port->prp[2] - port->bp);
	    break;
	}
	glSelectBuffer(1024, selectBuf);
	(void) glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(-1);
	glMatrixMode(GL_MODELVIEW);
	
	SmDisplayObjects(port, 1);
	nhits = glRenderMode(GL_RENDER);
	
	glMatrixMode(GL_PROJECTION);
	
	if (nhits == 0) {
	    sprintf(buf, "{%d %d %d %d}", pick[1], pick[2], pick[3], pick[4]);
	    return buf;
	}
	
	/* pick the nearest object */

	buf[0] = '\0';
	depth = selectBuf[1];
	front = selectBuf;
	hit = selectBuf + 3 + selectBuf[0];
	for (i = 1; i < nhits; i++) {
	    if (*hit != 1) {
		hit = hit + 3 + *hit;
		continue;
	    }
	    if (hit[1] < depth) {
		depth = hit[1];
		front = hit;
	    }
	    hit = hit + 3 + *hit;
	}
	
	for (obj = port->cell->objects; obj; obj = obj->next) {
	    if (obj->pickname == front[3]) {
		sprintf(buf, "{%d %d %d %d} %s %d",
			pick[1], pick[2], pick[3], pick[4], obj->label, front[1]);
	    }
	}
	return buf;
    } else {
	return "";
    }
}

/*
 *--------------------------------------------------------------
 *
 * PortLightSourceParseProc:
 *
 * process the '-light<x>' option for a viewport item.
 *
 * syntax:
 *
 * {ambient diffuse specular {x y z w} {dx dy dz} exponent cutoff {k0 k1 k2}}
 *
 * The first four parameters are required, others are optional.
 * If an optional parameter is not specified, default value will
 * be assigned instead:
 *
 * {dx dy dz}	= {0.0 0.0 -1.0}
 * exponent	= 0.0
 * cutoff	= 180.0
 * {k0 k1 k2}	= {1.0 0.0 0.0}
 *
 *--------------------------------------------------------------
 */

static int
PortLightSourceParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int index;
    int argc;
    char **argv;
    PortItem *port = (PortItem *) widgRec;
    LightSource *light;
    XColor *ambient, *diffuse, *specular;
    float pos[4], dir[3], att[3], ndir[3];
    double exponent, cutoff;

    index = (int) clientData;
    ambient = diffuse = specular = NULL;
    light = &port->lights[index];

    if (*value) {

	if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	    return TCL_ERROR;
	}
	if ((argc < 4) || (argc > 8)) {
	    Tcl_AppendResult(interp,
			     "wrong # args: should be {ambient diffuse specular position [direction exponent cutoff attenuation]}",
			     (char *) NULL);
	    free((void *) argv);
	    return TCL_ERROR;
	}

	/* get ambient, diffuse, and specular */
	if ((ambient = Tk_GetColor(interp, tkwin, Tk_GetUid(argv[0]))) == NULL) {
	    free((void *) argv);
	    return TCL_ERROR;
	}
	
	if ((diffuse = Tk_GetColor(interp, tkwin, Tk_GetUid(argv[1]))) == NULL) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    return TCL_ERROR;
	}
	
	if ((specular = Tk_GetColor(interp, tkwin, Tk_GetUid(argv[2]))) == NULL) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    Tk_FreeColor(diffuse);
	    return TCL_ERROR;
	}
	
	/* parse light position */
	if (ParseLightPosition(interp, argv[3], pos) != TCL_OK) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    Tk_FreeColor(diffuse);
	    Tk_FreeColor(specular);
	    return TCL_ERROR;
	}
	
	/* parse light direction */
	if ((argc > 4) && (ParseLightDirection(interp, argv[4], dir, ndir) != TCL_OK)) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    Tk_FreeColor(diffuse);
	    Tk_FreeColor(specular);
	    return TCL_ERROR;
	} else {
	    dir[0] = dir[1] = 0.0;
	    dir[2] = -1.0;
	}
	
	/* parse spotlight exponent */
	if ((argc > 5) && (Tcl_GetDouble(interp, argv[5], &exponent) != TCL_OK)) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    Tk_FreeColor(diffuse);
	    Tk_FreeColor(specular);
	    return TCL_ERROR;
	} else {
	    exponent = 0.0;
	}
	
	/* parse spotlight cutoff */
	if ((argc > 6) && (Tcl_GetDouble(interp, argv[6], &cutoff) != TCL_OK)) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    Tk_FreeColor(diffuse);
	    Tk_FreeColor(specular);
	    return TCL_ERROR;
	} else {
	    cutoff = 180.0;
	}
	
	/* parse light attenuation factors */
	if ((argc > 7) && (ParseLightAttenuation(interp, argv[7], att) != TCL_OK)) {
	    free((void *) argv);
	    Tk_FreeColor(ambient);
	    Tk_FreeColor(diffuse);
	    Tk_FreeColor(specular);
	    return TCL_ERROR;
	} else {
	    att[0] = 1.0;
	    att[1] = att[2] = 0.0;
	}
	free((void *) argv);
	
	/* light specification legal */
	if (light->ambientColor) Tk_FreeColor(light->ambientColor);
	if (light->diffuseColor) Tk_FreeColor(light->diffuseColor);
	if (light->specularColor) Tk_FreeColor(light->specularColor);
	light->ambientColor = ambient;
	light->diffuseColor = diffuse;
	light->specularColor = specular;
	light->position[0] = pos[0];
	light->position[1] = pos[1];
	light->position[2] = pos[2];
	light->position[3] = pos[3];
	light->direction[0] = dir[0];
	light->direction[1] = dir[1];
	light->direction[2] = dir[2];
	light->ndirection[0] = ndir[0];
	light->ndirection[1] = ndir[1];
	light->ndirection[2] = ndir[2];
	light->exponent = exponent;
	light->cutoff = cutoff;
	light->attenuation[0] = att[0];
	light->attenuation[1] = att[1];
	light->attenuation[2] = att[2];
	light->on = 1;
    }
    else {
	light->on = 0;
	if (light->ambientColor) Tk_FreeColor(light->ambientColor);
	if (light->diffuseColor) Tk_FreeColor(light->diffuseColor);
	if (light->specularColor) Tk_FreeColor(light->specularColor);
	light->ambientColor = light->diffuseColor = light->specularColor = NULL;
    }
    return TCL_OK;
}
/*
 *--------------------------------------------------------------
 *
 * ParseLightDirection:
 *
 * process light direction value.
 *
 *--------------------------------------------------------------
 */

static int
ParseLightDirection(interp, value, dir, ndir)
     Tcl_Interp *interp;
     char *value;
     float dir[];
     float ndir[];
{
    int argc;
    char **argv;
    double d[3], e;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: light direction should be {x y z}.", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }
    if (Tcl_GetDouble(interp, argv[0], &d[0]) != TCL_OK ||
	Tcl_GetDouble(interp, argv[1], &d[1]) != TCL_OK ||
	Tcl_GetDouble(interp, argv[2], &d[2]) != TCL_OK) {
	free((void *) argv);
	return TCL_ERROR;
    }
    dir[0] = (float) d[0];
    dir[1] = (float) d[1];
    dir[2] = (float) d[2];

    if ((e = d[0] * d[0] + d[1] * d[1] + d[2] * d[2]) > 0) {
	e  = 1 / e;
	ndir[0] = (float) (d[0] * e);
	ndir[1] = (float) (d[1] * e);
	ndir[2] = (float) (d[2] * e);
    } else {
	ndir[0] = ndir[1] = ndir[2] = 0;
    }
    free((void *) argv);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ParseLightPosition:
 *
 * process light position value.
 *
 * light position is given as {x y z w}. w must be >= 0. If
 * w == 0, light is directional, otherwise it is positional,
 * 
 *--------------------------------------------------------------
 */

static int
ParseLightPosition(interp, value, pos)
     Tcl_Interp *interp;
     char *value;
     float pos[];
{
    int argc;
    char **argv;
    double p0, p1, p2, p3;
    
    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc != 4) {
	Tcl_AppendResult(interp, "wrong # args: light position must be {x y z w}", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }
    if (Tcl_GetDouble(interp, argv[0], &p0) != TCL_OK ||
	Tcl_GetDouble(interp, argv[1], &p1) != TCL_OK ||
	Tcl_GetDouble(interp, argv[2], &p2) != TCL_OK ||
	Tcl_GetDouble(interp, argv[3], &p3) != TCL_OK) {
	free((void *) argv);
	return TCL_ERROR;
    }
    if (p3 < 0) {
	Tcl_AppendResult(interp, "bad value: w must be greater than or equal to 0", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }
    pos[0] = (float) p0;
    pos[1] = (float) p1;
    pos[2] = (float) p2;
    pos[3] = (float) p3;
    free((void *) argv);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ParseLightAttenuation:
 *
 * process light source attenuation factors.
 *
 *--------------------------------------------------------------
 */

static int
ParseLightAttenuation(interp, value, att)
     Tcl_Interp *interp;
     char *value;
     float att[];
{
    int argc;
    char **argv;
    double a0, a1, a2;

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: light source attenuation has 3 elements.", (char *) NULL);
	free((void *) argv);
	return TCL_ERROR;
    }
    if (Tcl_GetDouble(interp, argv[0], &a0) != TCL_OK ||
	Tcl_GetDouble(interp, argv[1], &a1) != TCL_OK ||
	Tcl_GetDouble(interp, argv[2], &a2) != TCL_OK) {
	free((void *) argv);
	return TCL_ERROR;
    }
    free((void *) argv);

    /* make sure factors have valid values */
    if ((a0 <= 0) || (a1 < 0) || (a2 < 0)) {
	Tcl_AppendResult(interp, "bad value: light source attenuation factors must be positive.", (char *) NULL);
	return TCL_ERROR;
    }
    att[0] = (float) a0;
    att[1] = (float) a1;
    att[2] = (float) a2;

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PortLightSourcePrintProc:
 *
 * produce a printable string for the '-light' option.
 *
 *--------------------------------------------------------------
 */

static char *
PortLightSourcePrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    char x[TCL_DOUBLE_SPACE], y[TCL_DOUBLE_SPACE], z[TCL_DOUBLE_SPACE], w[TCL_DOUBLE_SPACE];
    char dx[TCL_DOUBLE_SPACE], dy[TCL_DOUBLE_SPACE], dz[TCL_DOUBLE_SPACE];
    char exponent[TCL_DOUBLE_SPACE], cutoff[TCL_DOUBLE_SPACE];
    char k0[TCL_DOUBLE_SPACE], k1[TCL_DOUBLE_SPACE], k2[TCL_DOUBLE_SPACE];
    static char buf[256];

    PortItem *port = (PortItem *) widgRec;
    Tcl_Interp *interp = port->interp;
    LightSource *light = &port->lights[(int) clientData];

    *freeProcPtr = NULL;
    if (light->on) {
	Tcl_PrintDouble(interp, (double) light->position[0], x);
	Tcl_PrintDouble(interp, (double) light->position[1], y);
	Tcl_PrintDouble(interp, (double) light->position[2], z);
	Tcl_PrintDouble(interp, (double) light->position[3], w);

	Tcl_PrintDouble(interp, (double) light->direction[0], dx);
	Tcl_PrintDouble(interp, (double) light->direction[1], dy);
	Tcl_PrintDouble(interp, (double) light->direction[2], dz);

	Tcl_PrintDouble(interp, (double) light->attenuation[0], k0);
	Tcl_PrintDouble(interp, (double) light->attenuation[1], k1);
	Tcl_PrintDouble(interp, (double) light->attenuation[2], k2);

	Tcl_PrintDouble(interp, (double) light->cutoff, cutoff);
	Tcl_PrintDouble(interp, (double) light->exponent, exponent);

	sprintf(buf, "%s %s %s {%s %s %s %s} {%s %s %s} %s %s {%s %s %s}",
		Tk_NameOfColor(light->ambientColor),
		Tk_NameOfColor(light->diffuseColor),
		Tk_NameOfColor(light->specularColor),
		x, y, z, w,
		dx, dy, dz,
		exponent, cutoff, k0, k1, k2);
	return buf;
    }
    return "";
}

/*
 *--------------------------------------------------------------
 *
 * PortDepthCueParseProc:
 *
 * process the '-depthcue' option for a viewport item.
 *
 * syntax: color mode {density | start end}
 *
 *--------------------------------------------------------------
 */

static int
PortDepthCueParseProc(clientData, interp, tkwin, value, widgRec, offset)
     ClientData clientData;
     Tcl_Interp *interp;
     Tk_Window tkwin;
     char *value;
     char *widgRec;
     int offset;
{
    int argc;
    char **argv;
    XColor *color;
    double start, end, density;
    GLenum mode;
    PortItem *port = (PortItem *) widgRec;

    if (!(*value)) {
	if (port->depthcue.color) {
	    Tk_FreeColor(port->depthcue.color);
	}
	port->depthcue.color = NULL;
    }
    else {
	if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	    return TCL_ERROR;
	}
	if ((argc != 3) && (argc != 4)) {
	    Tcl_AppendResult(interp,
			     "wrong # args: should be {color mode density} or {color mode start end}",
			     (char *) NULL);
	    free((void *) argv);
	    return TCL_ERROR;
	}

	/* parse depth cue color */
	if (port->canvas) {
	    tkwin = Tk_CanvasTkwin(port->canvas);
	} else {
	    tkwin = port->tkwin;
	}
	if ((color = Tk_GetColor(interp, tkwin, Tk_GetUid(argv[0]))) == NULL) {
	    free((void *) argv);
	    return TCL_ERROR;
	}

	/* parse depth cue parameter values */
	if ((!strcmp(argv[1], "exp")) || (!strcmp(argv[1], "exp2"))) {
	    if (argc != 3) {
		Tcl_AppendResult(interp,
				 "wrong # args: should be {color ", argv[1], " density}",
				 (char *) NULL);
		free((void *) argv);
		Tk_FreeColor(color);
		return TCL_ERROR;
	    }
	    if ((argv[1])[3] == '2') {
		mode = GL_EXP2;
	    } else {
		mode = GL_EXP;
	    }
	    if (Tcl_GetDouble(interp, argv[2], &density) != TCL_OK) {
		Tk_FreeColor(color);
		free((void *) argv);
		return TCL_ERROR;
	    }
	} else if (!strcmp(argv[1], "linear")) {
	    if (argc != 4) {
		Tcl_AppendResult(interp,
				 "wrong # args: should be {color linear start end}",
				 (char *) NULL);
		free((void *) argv);
		Tk_FreeColor(color);
		return TCL_ERROR;
	    }
	    if ((Tcl_GetDouble(interp, argv[2], &start) != TCL_OK) ||
		(Tcl_GetDouble(interp, argv[3], &end) != TCL_OK)) {
		Tk_FreeColor(color);
		free((void *) argv);
		return TCL_ERROR;
	    }
	    mode = GL_LINEAR;
	}
	free((void *) argv);

	/* sanity check on values */

	if ((mode == GL_EXP) || (mode == GL_EXP2)) {
	    if (density < 0) {
		Tcl_AppendResult(interp,
				 "depth cue density must be non-negative",
				 (char *) NULL);
		Tk_FreeColor(color);
		return TCL_ERROR;
	    }
	} else if (mode == GL_LINEAR) {
	    if (start < 0) {
		Tcl_AppendResult(interp,
				 "depth cue parameter \"start\" must be non-negative",
				 (char *) NULL);
		Tk_FreeColor(color);
		return TCL_ERROR;
	    }
	    if (end < 0) {
		Tcl_AppendResult(interp,
				 "depth cue parameter \"end\" must be non-negative",
				 (char *) NULL);
		Tk_FreeColor(color);
		return TCL_ERROR;
	    }
	    if (end <= start) {
		Tcl_AppendResult(interp,
				 "depth cue parameter \"end\" must be greater than \"start\"",
				 (char *) NULL);
		Tk_FreeColor(color);
		return TCL_ERROR;
	    }
	}

	if (port->depthcue.color) {
	    Tk_FreeColor(port->depthcue.color);
	}
	port->depthcue.mode = mode;
	port->depthcue.color = color;
	port->depthcue.density = (DOUBLE) density;
	port->depthcue.start = (DOUBLE) start;
	port->depthcue.end = (DOUBLE) end;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PortDepthCuePrintProc:
 *
 * produce a printable string for the '-depthcue' option.
 *
 *--------------------------------------------------------------
 */

static char *
PortDepthCuePrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
     ClientData clientData;
     Tk_Window tkwin;
     char *widgRec;
     int offset;
     Tcl_FreeProc **freeProcPtr;
{
    PortItem *port = (PortItem *) widgRec;
    Tcl_Interp *interp = port->interp;
    DepthCue *cue = &port->depthcue;
    char density[TCL_DOUBLE_SPACE];
    char start[TCL_DOUBLE_SPACE], end[TCL_DOUBLE_SPACE];
    static char buf[256];

    *freeProcPtr = NULL;
    if (cue->color) {
	if (cue->mode == GL_EXP) {
	    Tcl_PrintDouble(interp, (double) cue->density, density);
	    sprintf(buf, "%s exp %s", Tk_NameOfColor(cue->color), density);
	} else if (cue->mode == GL_EXP2) {
	    Tcl_PrintDouble(interp, (double) cue->density, density);
	    sprintf(buf, "%s exp2 %s", Tk_NameOfColor(cue->color), density);
	} else if (cue->mode == GL_LINEAR) {
	    Tcl_PrintDouble(interp, (double) cue->start, start);
	    Tcl_PrintDouble(interp, (double) cue->end, end );
	    sprintf(buf, "%s linear %s %s", Tk_NameOfColor(cue->color), start, end);
	}
	return buf;
    }
    else {
	return "";
    }
}
